from typing import Any, Generator

import pytest
from _pytest.outcomes import fail
from sqlalchemy.orm import sessionmaker

from daomodel import DAOModel
from daomodel.dao import NotFound, DAO
from daomodel.db import DAOFactory, create_engine, init_db


class TestDAOFactory(DAOFactory):
    """
    A DAOFactory specifically designed for pytest.
    Includes functionality that can assert what is committed within the DB (through a secondary Session).
    """
    def __init__(self):
        engine = create_engine()  # create_engine('test.db')  # See DB for debugging, must delete file to rerun
        init_db(engine)
        super().__init__(sessionmaker(bind=engine))

    def __enter__(self) -> 'TestDAOFactory':
        super().__enter__()
        return self

    def assert_in_db(self, model: type[DAOModel], *pk, **expected_values: Any) -> None:
        """
        Assert that an object with specific attribute values is present in the DB.
        This checks the committed state of the database, not the session state.

        :param model: The DB table to check
        :param pk: The primary key values of the row
        :param expected_values: The column values to assert
        """
        with self.session_factory() as fresh_session:
            try:
                persisted_copy = DAO(model, fresh_session).get(*pk)
                for key, expected in expected_values.items():
                    actual = getattr(persisted_copy, key)
                    assert actual == expected, f'expected {key} of {persisted_copy} to be {expected} but was {actual}'
            except NotFound as e:
                fail(e.detail)

    def assert_not_in_db(self, model: type[DAOModel], *pk: Any) -> None:
        """
        Assert that the specified object is not present in the DB.
        This checks the committed state of the database, not the session state.

        :param model: The DB table to check
        :param pk: The primary key values of the row
        """
        with self.session_factory() as fresh_session:
            with pytest.raises(NotFound):
                DAO(model, fresh_session).get(*pk)

    def __exit__(self, exc_type, exc_value, traceback):
        super().__exit__(exc_type, exc_value, traceback)


@pytest.fixture(name='daos')
def daos_fixture() -> Generator[TestDAOFactory, Any, None]:
    """
    Provides a DAOFactory for Testing as a pytest fixture named `daos`.
    """
    with TestDAOFactory() as daos:
        yield daos



from datetime import timezone, datetime
from typing import Optional

import pytest

from daomodel import DAOModel
from daomodel.dao import DAO
from daomodel.fields import utc_now, Identifier, CurrentTimestampField, AutoUpdatingTimestampField
from tests.conftest import TestDAOFactory

ZERO_TIMESTAMP = datetime(1970, 1, 1)


class BasicModel(DAOModel, table=True):
    id: Identifier[int]


class TimestampsModel(DAOModel, table=True):
    id: Identifier[int]
    name: Optional[str]
    created_at: datetime = CurrentTimestampField
    updated_at: datetime = AutoUpdatingTimestampField


@pytest.fixture(name='timestamps_model_dao')
def timestamps_model_dao_fixture(daos: TestDAOFactory) -> DAO:
    return daos[TimestampsModel]


def test_utc_now():
    now = utc_now()
    assert isinstance(now, datetime)
    assert now.tzinfo is timezone.utc


def test_current_timestamp(timestamps_model_dao: DAO):
    timestamps_model_dao.create(1)
    entry = timestamps_model_dao.get(1)
    assert entry.created_at is not None
    assert entry.created_at > ZERO_TIMESTAMP


def test_auto_updating_timestamp(timestamps_model_dao: DAO):
    timestamps_model_dao.create(1)
    entry = timestamps_model_dao.get(1)
    updated_at = entry.updated_at
    assert updated_at is not None
    assert updated_at > ZERO_TIMESTAMP

    entry.name = 'Test'
    timestamps_model_dao.commit()
    assert entry.updated_at > updated_at



from typing import Optional, Any
from uuid import UUID

import pytest
from sqlalchemy.exc import IntegrityError
from sqlmodel import Field

from daomodel import DAOModel
from daomodel.fields import Protected, Identifier
from tests.conftest import TestDAOFactory
from tests.labeled_tests import labeled_tests
from tests.model_factory import create_test_model
from tests.test_fields import BasicModel
from tests.test_fields__inherited import get_test_cases


class OtherModel(DAOModel, table=True):
    id: Identifier[str]


class UUIDModel(DAOModel, table=True):
    id: Identifier[UUID]


class UUIDReferenceModel(DAOModel, table=True):
    id: Identifier[int]
    other_id: UUIDModel


class UUIDOptionalReferenceModel(DAOModel, table=True):
    id: Identifier[int]
    other_id: Optional[UUIDModel]


@labeled_tests({**get_test_cases('field')})
def test_field(annotation: Any):
    model_type = create_test_model(annotation)
    with TestDAOFactory() as daos:
        dao = daos[model_type]
        dao.create_with(id=1, value='test')
        entry = dao.find(id=1)
        assert entry.only().value == 'test'


@labeled_tests(get_test_cases('uuid'))
def test_uuid(annotation: Any):
    model_type = create_test_model(annotation)
    with TestDAOFactory() as daos:
        dao = daos[model_type]
        entry = dao.create_with(id=1)

        assert getattr(entry, 'value') is not None
        assert isinstance(getattr(entry, 'value'), UUID)

        entry2 = dao.create_with(id=2)
        assert getattr(entry2, 'value') is not None
        assert isinstance(getattr(entry2, 'value'), UUID)
        assert getattr(entry, 'value') != getattr(entry2, 'value')


@labeled_tests(get_test_cases('dict'))
def test_dict(annotation: Any):
    model_type = create_test_model(annotation)
    sample_dict = {'Hello': 'World!', 'Lorem': ['Ipsum', 'Dolor', 'Sit', 'Amet']}
    with TestDAOFactory() as daos:
        daos[model_type].create_with(id=1, value=sample_dict)
        daos.assert_in_db(model_type, 1, value=sample_dict)


@labeled_tests(get_test_cases('reference'))
def test_reference(annotation: Any):
    model_type = create_test_model(annotation)
    with TestDAOFactory() as daos:
        daos[BasicModel].create(100)
        dao = daos[model_type]
        dao.create_with(id=1, value=100)
        entry = dao.find(id=1)
        assert entry.only().value == 100


def test_reference__uuid(daos: TestDAOFactory):
    uuid_dao = daos[UUIDModel]
    reference_dao = daos[UUIDReferenceModel]

    uuid_model = daos[UUIDModel].create_with()
    reference_dao.create_with(id=1, other_id=uuid_model.id)

    entry = reference_dao.get(1)
    assert entry.other_id is not None
    assert uuid_dao.get(entry.other_id) is not None


def test_reference__uuid__optional(daos: TestDAOFactory):
    daos[UUIDOptionalReferenceModel].create(1)
    daos.assert_in_db(UUIDOptionalReferenceModel, 1, other_id=None)



from datetime import datetime, timezone
from typing import TypeVar, Generic, Any, Optional

from sqlmodel import Field
from sqlalchemy import Column
from sqlmodel.main import FieldInfo

from daomodel.util import reference_of

T = TypeVar('T')


class Identifier(Generic[T]):
    """A type annotation for primary key fields.

    Usage:
        class MyModel(DAOModel, table=True)
            id: Identifier[str]
            ...
    """
    pass


class Unsearchable(Generic[T]):
    """A type annotation to mark a field as not searchable.

    Usage:
        class MyModel(DAOModel, table=True)
            ...
            internal_notes: Unsearchable[str]
            ...
    """
    pass


class Protected(Generic[T]):
    """A type annotation for foreign key fields with RESTRICT delete behavior.

    This prevents the referenced object from being deleted if it is still referenced.

    Usage:
        class MyModel(DAOModel, table=True)
            ...
            parent: Protected[ParentModel]
            ...
    """
    pass


class ReferenceTo(FieldInfo):
    """Shortcut for defining a foreign key field.

    This class is used by the metaclass to set up foreign key constraints.
    It stores the target information, which the metaclass can then use
    to create the appropriate foreign key constraints and configurations.

    :param target: Either a string in the format 'table.column', a Column object, or a model attribute
    :param **kwargs: Additional arguments to pass to Field

    Usage:
        class MyModel(DAOModel, table=True)
            ...
            other_id: int = ReferenceTo('other_model.id')
            # or
            other_id: int = ReferenceTo(OtherModel.id)
    """
    def __init__(self, target: Optional[str|Column|Any] = None, **kwargs: Any):
        if 'foreign_key' not in kwargs:
            kwargs['foreign_key'] = (
                target if isinstance(target, str) else
                reference_of(target) if target is not None else
                None
            )
        super().__init__(**kwargs)


def utc_now():
    """Returns the current UTC time with timezone information."""
    return datetime.now(timezone.utc)


CurrentTimestampField = Field(default_factory=utc_now)
AutoUpdatingTimestampField = Field(default_factory=utc_now, sa_column_kwargs={'onupdate': utc_now})



from typing import Optional

import sqlalchemy
from sqlalchemy import Engine, event
from sqlalchemy.orm import sessionmaker
from sqlmodel import SQLModel

from daomodel import DAOModel
from daomodel.dao import DAO
from daomodel.transaction import TransactionMixin


def create_engine(path: Optional[str] = None) -> Engine:
    """Creates an SQLite Engine.

    :param path: the path to the DB file or None to keep the DB in-memory
    :return: The newly created SQLite Engine
    """
    if path is None:
        path = ''
        pool = sqlalchemy.StaticPool
    else:
        path = '/' + path
        pool = None
    return sqlalchemy.create_engine(
        'sqlite://' + path,
        connect_args={'check_same_thread': False},
        poolclass=pool
    )


@event.listens_for(Engine, 'connect')
def enforce_fk_constraints_for_sqlite(connection, _connection_record) -> None:
    cursor = connection.cursor()
    cursor.execute('pragma foreign_keys=on')
    cursor.close()


def init_db(engine: Engine) -> None:
    """Initiates DB tables of all imported SQL/DAOModels

    :param engine: The Engine for which to initialize the DB
    """
    SQLModel.metadata.create_all(engine)


class DAOFactory(TransactionMixin):
    """A Factory for creating DAOs for DAOModels.

    All DAOs/Sessions are auto closed when opened using a `with` statement.
    """
    def __init__(self, session_factory: sessionmaker):
        self.session_factory = session_factory

    def __enter__(self) -> 'DAOFactory':
        self.db = self.session_factory()
        return self

    def __getitem__(self, model: type[DAOModel]) -> DAO:
        return DAO(model, self.db)

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.db.close()



from decimal import Decimal
from enum import Enum, auto
from typing import Optional, Any

import pytest

from daomodel import DAOModel
from daomodel.fields import Identifier
from daomodel.model_diff import ModelDiff, Preference
from tests.labeled_tests import labeled_tests


class LaundryStatus(Enum):
    Public = auto()
    Shared = auto()
    Private = auto()


class Rental(DAOModel, table=True):
    address: Identifier[str]
    apt: Identifier[Optional[str]]
    dwelling_type: str
    sqft: int
    bedrooms: int
    bathrooms: Decimal
    garage_parking: int = 0
    laundry: Optional[LaundryStatus]
    cost: int


one_full = Decimal('1')
one_full_one_half = Decimal('1.1')
two_full = Decimal('2')
two_full_two_half = Decimal('2.2')


dorm = Rental(
    address='123 College Ave',
    apt='12',
    dwelling_type='Dormitory',
    sqft=200,
    bedrooms=1,
    bathrooms=one_full,
    laundry=LaundryStatus.Public,
    cost=0
)
apartment = Rental(
    address='456 City Plaza',
    apt='101',
    dwelling_type='Apartment',
    sqft=850,
    bedrooms=2,
    bathrooms=one_full,
    garage_parking=1,
    laundry=LaundryStatus.Shared,
    cost=1200
)
apartment_two = Rental(
    address='456 City Plaza',
    apt='102',
    dwelling_type='Apartment',
    sqft=850,
    bedrooms=2,
    bathrooms=one_full,
    garage_parking=1,
    laundry=LaundryStatus.Shared,
    cost=1200
)
multi_family = Rental(
    address='789 Suburb St',
    apt='B',
    dwelling_type='Multi-family House',
    sqft=950,
    bedrooms=1,
    bathrooms=one_full_one_half,
    cost=1800
)
town_home = Rental(
    address='321 Maple Dr',
    dwelling_type='Town home',
    sqft=1400,
    bedrooms=3,
    bathrooms=two_full,
    garage_parking=1,
    laundry=LaundryStatus.Private,
    cost=2200
)
single_family = Rental(
    address='987 Country Rd',
    dwelling_type='House',
    sqft=1800,
    bedrooms=4,
    bathrooms=two_full_two_half,
    garage_parking=2,
    laundry=LaundryStatus.Private,
    cost=3000
)


class RentalDiff(ModelDiff):
    def get_preferred(self, field: str) -> Preference:
        match field:
            case 'address'|'apt':
                return Preference.NOT_APPLICABLE
            case 'sqft'|'bedrooms'|'bathrooms'|'garage_parking':
                return Preference.LEFT if self.get_left(field) > self.get_right(field) else Preference.RIGHT
            case 'cost':
                return Preference.LEFT if self.get_left(field) < self.get_right(field) else Preference.RIGHT
            case 'dwelling_type':
                bad = ['Apartment', 'Multi-family House']
                good = ['Town home', 'House']
                def get_value(dwelling_type: str):
                    if dwelling_type in bad:
                        return -1
                    elif dwelling_type in good:
                        return 1
                    else:
                        return 0
                left_value = get_value(self.get_left(field))
                right_value = get_value(self.get_right(field))
                return(
                    Preference.LEFT if left_value > right_value else
                    Preference.RIGHT if right_value > left_value else
                    Preference.BOTH if left_value > 0 else
                    Preference.NEITHER
                )
            case 'laundry':
                left_value = 0 if self.get_left(field) is None else self.get_left(field).value
                right_value = 0 if self.get_right(field) is None else self.get_right(field).value
                return Preference.LEFT if left_value > right_value else Preference.RIGHT
            case _:
                raise NotImplementedError(f'The field {field} is not supported')


@labeled_tests({
    'dorm vs apartment':
        (dorm, apartment, {
            'dwelling_type': ('Dormitory', 'Apartment'),
            'sqft': (200, 850),
            'bedrooms': (1, 2),
            'garage_parking': (0, 1),
            'laundry': (LaundryStatus.Public, LaundryStatus.Shared),
            'cost': (0, 1200)
        }),
    'dorm vs multi family':
        (dorm, multi_family, {
            'dwelling_type': ('Dormitory', 'Multi-family House'),
            'sqft': (200, 950),
            'bathrooms': (Decimal('1'), Decimal('1.1')),
            'laundry': (LaundryStatus.Public, None),
            'cost': (0, 1800)
        }),
    'dorm vs town home':
        (dorm, town_home, {
            'dwelling_type': ('Dormitory', 'Town home'),
            'sqft': (200, 1400),
            'bedrooms': (1, 3),
            'bathrooms': (Decimal('1'), Decimal('2')),
            'garage_parking': (0, 1),
            'laundry': (LaundryStatus.Public, LaundryStatus.Private),
            'cost': (0, 2200)
        }),
    'dorm vs single family':
        (dorm, single_family, {
            'dwelling_type': ('Dormitory', 'House'),
            'sqft': (200, 1800),
            'bedrooms': (1, 4),
            'bathrooms': (Decimal('1'), Decimal('2.2')),
            'garage_parking': (0, 2),
            'laundry': (LaundryStatus.Public, LaundryStatus.Private),
            'cost': (0, 3000)
        }),
    'apartment vs multi family':
        (apartment, multi_family, {
            'dwelling_type': ('Apartment', 'Multi-family House'),
            'sqft': (850, 950),
            'bedrooms': (2, 1),
            'bathrooms': (Decimal('1'), Decimal('1.1')),
            'garage_parking': (1, 0),
            'laundry': (LaundryStatus.Shared, None),
            'cost': (1200, 1800)
        }),
    'apartment vs town home':
        (apartment, town_home, {
            'dwelling_type': ('Apartment', 'Town home'),
            'sqft': (850, 1400),
            'bedrooms': (2, 3),
            'bathrooms': (Decimal('1'), Decimal('2')),
            'laundry': (LaundryStatus.Shared, LaundryStatus.Private),
            'cost': (1200, 2200)
        }),
    'apartment vs single family':
        (apartment, single_family, {
            'dwelling_type': ('Apartment', 'House'),
            'sqft': (850, 1800),
            'bedrooms': (2, 4),
            'bathrooms': (Decimal('1'), Decimal('2.2')),
            'garage_parking': (1, 2),
            'laundry': (LaundryStatus.Shared, LaundryStatus.Private),
            'cost': (1200, 3000)
        }),
    'multi family vs town home':
        (multi_family, town_home, {
            'dwelling_type': ('Multi-family House', 'Town home'),
            'sqft': (950, 1400),
            'bedrooms': (1, 3),
            'bathrooms': (Decimal('1.1'), Decimal('2')),
            'garage_parking': (0, 1),
            'laundry': (None, LaundryStatus.Private),
            'cost': (1800, 2200)
        }),
    'multi family vs single family':
        (multi_family, single_family, {
            'dwelling_type': ('Multi-family House', 'House'),
            'sqft': (950, 1800),
            'bedrooms': (1, 4),
            'bathrooms': (Decimal('1.1'), Decimal('2.2')),
            'garage_parking': (0, 2),
            'laundry': (None, LaundryStatus.Private),
            'cost': (1800, 3000)
        }),
    'town home vs single family':
        (town_home, single_family, {
            'dwelling_type': ('Town home', 'House'),
            'sqft': (1400, 1800),
            'bedrooms': (3, 4),
            'bathrooms': (Decimal('2'), Decimal('2.2')),
            'garage_parking': (1, 2),
            'cost': (2200, 3000)
        }),
    'empty diff':
        (apartment, apartment_two, {})
})
def test_model_diff(left: Rental, right: Rental, expected: dict[str, tuple[Any, Any]]):
    assert ModelDiff(left, right, include_pk=False) == expected


@labeled_tests({
    'same address':
        (apartment, apartment_two, {
            'apt': ('101', '102')
        }),
    'right None':
        (multi_family, town_home, {
            'address': ('789 Suburb St', '321 Maple Dr'),
            'apt': ('B', None)
        }),
    'left None':
        (single_family, multi_family, {
            'address': ('987 Country Rd', '789 Suburb St'),
            'apt': (None, 'B')
        }),
    'both None':
        (town_home, single_family, {
            'address': ('321 Maple Dr', '987 Country Rd')
        })
})
def test_model_diff__pk(left: Rental, right: Rental, expected: dict[str, tuple[Any, Any]]):
    diff = ModelDiff(left, right)
    pk_diff = {key: diff[key] for key in ['address', 'apt'] if key in diff}
    assert pk_diff == expected


def test_get_left_get_right():
    diff = ModelDiff(apartment, apartment_two)
    assert diff.get_left('apt') == '101'
    assert diff.get_right('apt') == '102'


def test_get_left_get_right__missing():
    diff = ModelDiff(apartment, apartment_two)
    with pytest.raises(KeyError):
        diff.get_left('address')
    with pytest.raises(KeyError):
        diff.get_right('cost')


def test_get_left_get_right__invalid():
    diff = ModelDiff(apartment, apartment_two)
    with pytest.raises(KeyError):
        diff.get_left('APT')
    with pytest.raises(KeyError):
        diff.get_right('invalid')


@labeled_tests({
    'dorm vs apartment': [
        (dorm, apartment, 'dwelling_type', Preference.LEFT),
        (dorm, apartment, 'sqft', Preference.RIGHT),
        (dorm, apartment, 'bedrooms', Preference.RIGHT),
        (dorm, apartment, 'garage_parking', Preference.RIGHT),
        (dorm, apartment, 'laundry', Preference.RIGHT),
        (dorm, apartment, 'cost', Preference.LEFT)
    ],
    'dorm vs multi family': [
        (dorm, multi_family, 'dwelling_type', Preference.LEFT),
        (dorm, multi_family, 'sqft', Preference.RIGHT),
        (dorm, multi_family, 'bathrooms', Preference.RIGHT),
        (dorm, multi_family, 'laundry', Preference.LEFT),
        (dorm, multi_family, 'cost', Preference.LEFT)
    ],
    'dorm vs town home': [
        (dorm, town_home, 'dwelling_type', Preference.RIGHT),
        (dorm, town_home, 'sqft', Preference.RIGHT),
        (dorm, town_home, 'bedrooms', Preference.RIGHT),
        (dorm, town_home, 'bathrooms', Preference.RIGHT),
        (dorm, town_home, 'garage_parking', Preference.RIGHT),
        (dorm, town_home, 'laundry', Preference.RIGHT),
        (dorm, town_home, 'cost', Preference.LEFT)
    ],
    'dorm vs single family': [
        (dorm, single_family, 'dwelling_type', Preference.RIGHT),
        (dorm, single_family, 'sqft', Preference.RIGHT),
        (dorm, single_family, 'bedrooms', Preference.RIGHT),
        (dorm, single_family, 'bathrooms', Preference.RIGHT),
        (dorm, single_family, 'garage_parking', Preference.RIGHT),
        (dorm, single_family, 'laundry', Preference.RIGHT),
        (dorm, single_family, 'cost', Preference.LEFT)
    ],
    'apartment vs multi family': [
        (apartment, multi_family, 'dwelling_type', Preference.NEITHER),
        (apartment, multi_family, 'sqft', Preference.RIGHT),
        (apartment, multi_family, 'bedrooms', Preference.LEFT),
        (apartment, multi_family, 'bathrooms', Preference.RIGHT),
        (apartment, multi_family, 'garage_parking', Preference.LEFT),
        (apartment, multi_family, 'laundry', Preference.LEFT),
        (apartment, multi_family, 'cost', Preference.LEFT)
    ],
    'apartment vs town home': [
        (apartment, town_home, 'dwelling_type', Preference.RIGHT),
        (apartment, town_home, 'sqft', Preference.RIGHT),
        (apartment, town_home, 'bedrooms', Preference.RIGHT),
        (apartment, town_home, 'bathrooms', Preference.RIGHT),
        (apartment, town_home, 'laundry', Preference.RIGHT),
        (apartment, town_home, 'cost', Preference.LEFT)
    ],
    'apartment vs single family': [
        (apartment, single_family, 'dwelling_type', Preference.RIGHT),
        (apartment, single_family, 'sqft', Preference.RIGHT),
        (apartment, single_family, 'bedrooms', Preference.RIGHT),
        (apartment, single_family, 'bathrooms', Preference.RIGHT),
        (apartment, single_family, 'garage_parking', Preference.RIGHT),
        (apartment, single_family, 'laundry', Preference.RIGHT),
        (apartment, single_family, 'cost', Preference.LEFT)
    ],
    'multi family vs town home': [
        (multi_family, town_home, 'dwelling_type', Preference.RIGHT),
        (multi_family, town_home, 'sqft', Preference.RIGHT),
        (multi_family, town_home, 'bedrooms', Preference.RIGHT),
        (multi_family, town_home, 'bathrooms', Preference.RIGHT),
        (multi_family, town_home, 'garage_parking', Preference.RIGHT),
        (multi_family, town_home, 'laundry', Preference.RIGHT),
        (multi_family, town_home, 'cost', Preference.LEFT)
    ],
    'multi family vs single family': [
        (multi_family, single_family, 'dwelling_type', Preference.RIGHT),
        (multi_family, single_family, 'sqft', Preference.RIGHT),
        (multi_family, single_family, 'bedrooms', Preference.RIGHT),
        (multi_family, single_family, 'bathrooms', Preference.RIGHT),
        (multi_family, single_family, 'garage_parking', Preference.RIGHT),
        (multi_family, single_family, 'laundry', Preference.RIGHT),
        (multi_family, single_family, 'cost', Preference.LEFT)
    ],
    'town home vs single family': [
        (town_home, single_family, 'dwelling_type', Preference.BOTH),
        (town_home, single_family, 'sqft', Preference.RIGHT),
        (town_home, single_family, 'bedrooms', Preference.RIGHT),
        (town_home, single_family, 'bathrooms', Preference.RIGHT),
        (town_home, single_family, 'garage_parking', Preference.RIGHT),
        (town_home, single_family, 'cost', Preference.LEFT)
    ],
    'n/a': [
        (multi_family, single_family, 'address', Preference.NOT_APPLICABLE),
        (multi_family, single_family, 'apt', Preference.NOT_APPLICABLE),
    ],
})
def test_get_preferred(left: Rental, right: Rental, field: str, expected: Preference):
    assert RentalDiff(left, right).get_preferred(field) == expected


def test_get_preferred__not_implemented():
    with pytest.raises(NotImplementedError):
        ModelDiff(apartment, multi_family).get_preferred('cost')


def test_get_preferred__missing():
    with pytest.raises(KeyError):
        RentalDiff(single_family, town_home).get_preferred('laundry')



from typing import Any
import uuid

from daomodel import DAOModel
from daomodel.fields import Identifier


def create_test_model(field_type: Any = None, field_name: str = 'value', base_model: type = None, inherited: bool = False) -> type[DAOModel]:
    """Dynamically creates a model with the specified field type.

    :param field_type: The type annotation for the field, or exclude to only have an id field
    :param field_name: The name of the field (default: 'value')
    :param base_model: Optional base model to inherit from
    :param inherited: True if the type field should belong to a parent model that is extended by the returned model
    :return: A dynamically created DAOModel class
    """
    annotations = {'id': Identifier[int]}
    if field_type is not None:
        annotations[field_name] = field_type

    model = type(
        f'DynamicModel{uuid.uuid4().hex[:8].capitalize()}',
        (base_model or DAOModel,),
        {
            '__annotations__': annotations,
            '__module__': 'tests.field_tests.model_factory'
        },
        table=not inherited
    )
    if inherited:
        model = create_test_model(str, field_name='child_field', base_model=model)

    return model



from abc import abstractmethod
from enum import Enum
from typing import Any, Optional, Callable

from daomodel import DAOModel
from daomodel.dao import Conflict
from daomodel.property_filter import DEFAULT


class Preference(Enum):
    """Represents the preference of values in a model comparison.

    The available options are:
    - LEFT: Indicates preference for the 'left' side value.
    - RIGHT: Indicates preference for the 'right' side value.
    - NEITHER: Indicates that neither value is preferred.
    - BOTH: Indicates that both values are equally preferable.
    - NOT_APPLICABLE: Indicates that a preference does not apply in the given context i.e. the name field of two customers.
    """
    NOT_APPLICABLE = -1
    NEITHER = 0
    LEFT = 1
    RIGHT = 2
    BOTH = 3

BASELINE_VALUE = Preference.LEFT
TARGET_VALUE = Preference.RIGHT


class ModelDiff(dict[str, tuple[Any, Any]]):
    """A dictionary wrapper to provide extended functionality for model comparisons."""
    def __init__(self, left: DAOModel, right: DAOModel, include_pk: Optional[bool] = True):
        super().__init__()
        self.left = left
        self.right = right
        self.update(left.compare(right, include_pk=include_pk))

    def get_left(self, field: str) -> Any:
        """Fetches the value of the left model.

        :param field: The name of the field to fetch
        :return: The left value for the specified field
        :raises KeyError: if the field is invalid or otherwise not included in this diff
        """
        if field not in self:
            raise KeyError(f'{field} not found in diff.')
        return self.get(field)[0]

    def get_right(self, field: str) -> Any:
        """Fetches the value of the right model.

        :param field: The name of the field to fetch
        :return: The right value for the specified field
        :raises KeyError: if the field is invalid or otherwise not included in this diff
        """
        if field not in self:
            raise KeyError(f'{field} not found in diff.')
        return self.get(field)[1]

    @abstractmethod
    def get_preferred(self, field: str) -> Preference:
        """Defines which of the varying values is preferred.

        :param field: The name of the field
        :return: The Preference between the possible values
        :raises KeyError: if the field is invalid or otherwise not included in this diff
        """
        raise NotImplementedError(f'Cannot determine which value is preferred for {field}: '
                                  f'{self.get_left(field)} -> {self.get_right(field)}')


class Unresolved:
    """Represents an unresolved conflict within a ChangeSet.

    This is a simple wrapper for the target value to act as a label.
    """
    def __init__(self, target: Any):
        self.target = target

    def __eq__(self, other: Any) -> bool:
        return self.target == other.target if isinstance(other, Unresolved) else False

    def __hash__(self) -> int:
        return hash(self.target)
    
    def __repr__(self) -> str:
        """Provide a meaningful string representation of the Unresolved instance."""
        return f'Unresolved(target={repr(self.target)})'


class Resolved:
    """Represents a resolved value for a specific conflict.

    Stores the original target as well as the resolution of a conflict in the context of a ChangeSet.
    """
    def __init__(self, target: Any, resolution: Any):
        self.target = target
        self.resolution = resolution

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Resolved):
            return self.target == other.target and self.resolution == other.resolution
        return False

    def __hash__(self) -> int:
        return hash((self.target, self.resolution))

    def __repr__(self) -> str:
        """Provide a meaningful string representation of the Unresolved instance."""
        return f'Resolved(target={repr(self.target)}, resolution={repr(self.resolution)})'


class ChangeSet(ModelDiff):
    """A directional model diff with the left being the baseline and the right being the target.

    Unlike the standard ModelDiff, PK is excluded by default.
    
    The `conflict_resolution` argument allows users to define specific conflict resolution rules
    for handling differences between the baseline and target models. This parameter can 
    accept field-specific resolution methods or a default resolution method for all fields.
    If no resolution can be determined, a `Conflict` exception is raised. Note: This resolution
    is only applicable for a Preference of `BOTH` as determined by `get_preferred()`.
    
    Examples:
    1. Field-Specific Resolution:
       ```
       def prefer_longer_name(values: list[str]):
           return max(values, key=len)
    
       ChangeSet(baseline, target, name=prefer_longer_name)
       ```
    
    2. Default Resolution:
       ```
       def conflict_str(values: list[Any]):
           return f'Unresolved conflict: {values}' 
       
       ChangeSet(baseline, target, default=conflict_str)
       ```
    
    3. Mixed Resolution:
       ```
       ChangeSet(baseline, target, default=conflict_str, amount=sum)
       ```
    
    4. Using a Preference for Conflict Resolution:
       ```
       from model_diff import Preference
       
       ChangeSet(baseline, target, default=Preference.LEFT, status=Preference.RIGHT)
       ```
       
       or use the helper constants:
       ```
       from model_diff import BASELINE_VALUE, TARGET_VALUE
       
       ChangeSet(baseline, target, default=BASELINE_VALUE, status=TARGET_VALUE)
       ```

    5. Static Resolution:
       ```
       ChangeSet(baseline, target, name='Static Name', default='Default Value')
       ```
    
    View the test code for more examples.
    """
    def __init__(self,
                 baseline: DAOModel,
                 target: DAOModel,
                 include_pk: Optional[bool] = False,
                 **conflict_resolution: Preference|Callable|Any):
        super().__init__(baseline, target, include_pk)
        self.modified_in_baseline = self.left.get_property_names(~DEFAULT)
        self.modified_in_target = self.right.get_property_names(~DEFAULT)
        self.conflict_resolution = conflict_resolution

    def get_baseline(self, field: str) -> Any:
        """Fetches the value of the baseline model.

        :param field: The name of the field to fetch
        :return: The baseline value for the specified field
        :raises KeyError: if the field is invalid or otherwise not included in this diff
        """
        return self.get_left(field)

    def get_target(self, field: str) -> Any:
        """Fetches the value of the target model.

        :param field: The name of the field to fetch
        :return: The target value for the specified field
        :raises KeyError: if the field is invalid or otherwise not included in this diff
        """
        return self.get_right(field)

    def has_target_value(self, field: str) -> bool:
        """Returns True if the target value for the specified field exists"""
        return self.get_target(field) is not None

    def all_values(self, field: str) -> list[Any]:
        """Returns a list containing the baseline value followed by the target value for the specified field."""
        return [self.get_left(field), self.get_right(field)]

    def get_resolution(self, field: str) -> Any:
        """Returns the resolved value for the specified field.

        This will be the new value for the specified field if the change set were to be applied.

        :param field: The name of the field to fetch
        :return: The resolved value, which is the target value unless resolve_preferences() was called
        """
        target = self.get_target(field)
        return target.resolution if isinstance(target, Resolved) else target

    def get_preferred(self, field: str) -> Preference:
        return (
            Preference.LEFT if not self.has_target_value(field) else
            Preference.BOTH if field in self.modified_in_baseline and field in self.modified_in_target else
            Preference.LEFT if field in self.modified_in_baseline else
            Preference.RIGHT if field in self.modified_in_target else
            Preference.NEITHER
        )

    def resolve_conflict(self, field: str) -> Preference|Any:
        """Defines how to handle conflicts between preferred values.

        A conflict occurs when both the baseline and target have unique meaningful values for a field.
        Conflict resolution guidelines are to be provided during construction if needed.
        Alternatively, the implementer may completely override this function in order to have full control.

        :param field: The field having a conflict
        :return: The result of the resolution which may be the baseline value, target value, or something new entirely
        :raises Conflict: if a resolution cannot be determined
        """
        def raise_conflict(*values: Any) -> Any:
            raise Conflict(msg=f'Unable to determine preferred result for {field}: {values}')

        default = self.conflict_resolution.get('default', raise_conflict)
        resolution_method = self.conflict_resolution.get(field, default)
        resolution = resolution_method(self.all_values(field)) if callable(resolution_method) else resolution_method

        return (
            Preference.LEFT if resolution == self.get_baseline(field) else
            Preference.RIGHT if resolution == self.get_target(field) else
            Preference.NEITHER if resolution is None else
            resolution
        )

    def resolve_preferences(self) -> 'ChangeSet':
        """Removes unwanted changes, preserving the meaningful values, regardless of them being from baseline or target

        :return: This ChangeSet to allow for chaining function calls
        :raises Conflict: if both baseline and target have meaningful values (unless resolve_conflict is overridden)
        """
        for field in list(self.keys()):
            preferred = self.get_preferred(field)
            if preferred == Preference.BOTH:
                preferred = self.resolve_conflict(field)
            match preferred:
                case Preference.NOT_APPLICABLE | Preference.NEITHER:
                    self[field] = (self.get_baseline(field), Unresolved(self.get_target(field)))
                case Preference.LEFT:
                    del self[field]
                case Preference.RIGHT:
                    pass
                case _:
                    self[field] = (self.get_baseline(field), Resolved(self.get_target(field), preferred))
        return self

    def apply(self) -> DAOModel:
        """Enacts these changes upon the baseline.

        You will typically want to call resolve_preferences prior to this.
        """
        self.left.set_values(**{field: self.get_resolution(field) for field in self.keys()})
        return self.left


class MergeSet(ChangeSet):
    """Used for managing and resolving differences between a baseline and multiple target models.

    This class extends functionality of `ChangeSet` to handle scenarios where
    multiple target models are compared against a baseline. It initializes with
    a baseline model and one or more target models, identifying differences and
    allowing for conflict resolution between them. The class provides utilities
    to access and resolve values for specific fields across the baseline and
    targets.

    For more details regarding initialization, see ChangeSet.
    """
    def __init__(self, baseline: DAOModel, *targets: DAOModel, **conflict_resolution: Preference|Callable|Any):
        super().__init__(baseline, targets[0], **conflict_resolution)
        self.left = baseline
        self.right = targets
        for model in targets:
            self.modified_in_target += model.get_property_names(~DEFAULT)

        for model in targets:
            for k, v in baseline.compare(model).items():
                self[k] = (v[0], [None] * len(targets))

        for index, model in enumerate(targets):
            for k, v in self.items():
                v[1][index] = model.get_value_of(k)

    def has_target_value(self, field: str) -> bool:
        """Returns whether the target contains any non-None values."""
        return any(target is not None for target in self.get_target(field))

    def all_values(self, field: str) -> list[Any]:
        """Returns a list containing the baseline value followed by all target values for the specified field."""
        return [self.get_baseline(field)] + self.get_target(field)

    def resolve_conflict(self, field: str) -> tuple[Preference.RIGHT, int]|Any:
        resolution = super().resolve_conflict(field)
        if resolution is Preference.NEITHER:
            resolution = None

        target_values = self.get_target(field)
        if resolution in target_values:
            if len(target_values) == 1:
                return Preference.RIGHT
            else:
                return Preference.RIGHT, target_values.index(resolution)
        elif resolution is None:
            return Preference.NEITHER
        else:
            return resolution



from typing import Optional

import pytest
from sqlalchemy.exc import IntegrityError
from sqlmodel import Field

from daomodel import DAOModel
from daomodel.fields import Identifier, Protected, ReferenceTo
from tests.conftest import TestDAOFactory
from tests.labeled_tests import labeled_tests
from tests.test_fields__types import OtherModel


class StandardReferenceModel(DAOModel, table=True):
    id: Identifier[int]
    other_id: OtherModel


class StandardReferenceToModel(DAOModel, table=True):
    id: Identifier[int]
    other_id: int = ReferenceTo('other_model.id')


class OptionalReferenceModel(DAOModel, table=True):
    id: Identifier[int]
    other_id: Optional[OtherModel]


class OptionalReferenceToModel(DAOModel, table=True):
    id: Identifier[int]
    other_id: Optional[int] = ReferenceTo('other_model.id')


class ProtectedReferenceModel(DAOModel, table=True):
    id: Identifier[int]
    other_id: Protected[OtherModel]


class ProtectedReferenceToModel(DAOModel, table=True):
    id: Identifier[int]
    other_id: Protected[int] = ReferenceTo('other_model.id')


class CustomReferenceModel(DAOModel, table=True):
    id: Identifier[int]
    other_id: Optional[OtherModel] = Field(foreign_key='auto', ondelete='CASCADE')


class CustomReferenceToModel(DAOModel, table=True):
    id: Identifier[int]
    other_id: Optional[int] = ReferenceTo('other_model.id', ondelete='CASCADE')


class ReferenceToColumnModel(DAOModel, table=True):
    id: Identifier[int]
    other_id: int = ReferenceTo(OtherModel.id)


class CircularReferenceModel(DAOModel, table=True):
    id: Identifier[int]
    circular: Optional[int] = ReferenceTo('circular_reference_model.id')


class NoTargetReferenceToModel(DAOModel, table=True):
    id: Identifier[int]
    other_id: Optional[OtherModel] = ReferenceTo(ondelete='CASCADE')


class ModelWithUniqueField(DAOModel, table=True):
    id: Identifier[int]
    unique: str = Field(unique=True)


class ReferenceToFieldModel(DAOModel, table=True):
    id: Identifier[int]
    ref: str = ReferenceTo(ModelWithUniqueField.unique)


def test_reference_to(daos: TestDAOFactory):
    daos[OtherModel].create('A')
    daos[StandardReferenceToModel].create_with(id=2, other_id='A')
    daos.assert_in_db(StandardReferenceToModel, 2, other_id='A')


@labeled_tests({
    'reference': StandardReferenceModel,
    'optional reference': OptionalReferenceModel,
    'reference to': StandardReferenceToModel,
    'optional reference to': OptionalReferenceToModel
})
def test_reference__cascade_on_update(model: type[DAOModel]):
    with TestDAOFactory() as daos:
        test_dao = daos[OtherModel]
        fk_dao = daos[model]

        other_entry = test_dao.create('A')
        fk_entry = fk_dao.create_with(id=1, other_id='A')

        test_dao.rename(other_entry, 'B')
        assert fk_entry.other_id == 'B'


@labeled_tests({
    'reference': StandardReferenceModel,
    'optional reference': OptionalReferenceModel,
    'reference to': StandardReferenceToModel,
    'optional reference to': OptionalReferenceToModel
})
def test_reference__cascade_on_delete(model: type[DAOModel]):
    with TestDAOFactory() as daos:
        test_dao = daos[OtherModel]
        fk_dao = daos[model]

        other_entry = test_dao.create('A')
        fk_dao.create_with(id=1, other_id='A')
        daos.assert_in_db(model, 1, other_id='A')

        test_dao.remove(other_entry)
        if model.doc_name().startswith('Optional'):
            daos.assert_in_db(model, 1, other_id=None)
        else:
            daos.assert_not_in_db(model, 1)


@labeled_tests({
    'reference': ProtectedReferenceModel,
    'reference to': ProtectedReferenceToModel
})
def test_reference__protected(model: type[DAOModel]):
    with TestDAOFactory() as daos:
        test_dao = daos[OtherModel]
        fk_dao = daos[model]

        other_entry = test_dao.create('A')
        fk_dao.create_with(id=1, other_id='A')

        with pytest.raises(IntegrityError):
            test_dao.remove(other_entry)


@labeled_tests({
    'reference': CustomReferenceModel,
    'reference to': CustomReferenceToModel
})
def test_reference__override_on_delete(model: type[DAOModel]):
    with TestDAOFactory() as daos:
        test_dao = daos[OtherModel]
        fk_dao = daos[model]

        fk_entry = fk_dao.create(1)
        daos.assert_in_db(model, 1, other_id=None)

        other_entry = test_dao.create('A')
        fk_entry.other_id = other_entry.id
        fk_dao.upsert(fk_entry)
        daos.assert_in_db(model, 1, other_id='A')

        test_dao.remove(other_entry)
        daos.assert_not_in_db(model, 1)


def test_reference_to__column(daos: TestDAOFactory):
    daos[OtherModel].create('A')
    daos[ReferenceToColumnModel].create_with(id=2, other_id='A')
    daos.assert_in_db(ReferenceToColumnModel, 2, other_id='A')


def test_reference_to__self(daos: TestDAOFactory):
    daos[CircularReferenceModel].create(1)
    daos[CircularReferenceModel].create_with(id=2, circular=1)
    daos.assert_in_db(CircularReferenceModel, 2, circular=1)


def test_reference_to__no_target(daos: TestDAOFactory):
    other = daos[OtherModel].create('A')
    daos[NoTargetReferenceToModel].create_with(id=1, other_id='A')
    daos.assert_in_db(NoTargetReferenceToModel, 1, other_id='A')

    daos[OtherModel].remove(other)
    daos.assert_not_in_db(NoTargetReferenceToModel, 1)


def test_reference_to__unique_field(daos: TestDAOFactory):
    daos[ModelWithUniqueField].create_with(id=0, unique='value')
    daos[ReferenceToFieldModel].create_with(id=1, ref='value')
    daos.assert_in_db(ReferenceToFieldModel, 1, ref='value')



from __future__ import annotations

from typing import Optional, Any, TypeVar, Iterable, Iterator

from sqlalchemy import func, Column, text, UnaryExpression
from sqlalchemy.orm import Session
from sqlalchemy.orm.query import Query

from daomodel.util import values_from_dict, retain_in_dict, MissingInput, InvalidArgumentCount, ensure_iter, dedupe, ConditionOperator
from daomodel.transaction import TransactionMixin, Conflict

from daomodel import DAOModel


class NotFound(Exception):
    """Indicates that the requested object could not be found."""
    def __init__(self, model: DAOModel):
        self.detail = f'{model.__class__.doc_name()} {model} not found'


class PrimaryKeyConflict(Conflict):
    """Indicates that the store could not be updated due to a primary key conflict."""
    def __init__(self, model: DAOModel):
        super().__init__(model, f'{model.__class__.doc_name()} {model} already exists')


Model = TypeVar('Model', bound=DAOModel)
class SearchResults(list[Model]):
    """The paginated results of a filtered search."""
    def __init__(self, results: list[Model], total: int = None, page: Optional[int] = None, per_page: Optional[int] = None):
        super().__init__(results)
        self.results = results
        self.total = len(results) if total is None else total
        self.page = page
        self.per_page = per_page

    def __iter__(self) -> Iterator[Model]:
        return iter(self.results)

    def __eq__(self, other: 'SearchResults') -> bool:
        return (self.results == other.results
                and self.total == other.total
                and self.page == other.page
                and self.per_page == other.per_page
                ) if type(self) == type(other) else False

    def __hash__(self) -> int:
        return hash((tuple(self.results), self.total, self.page, self.per_page))

    def __str__(self) -> str:
        string = str(self.results)
        if self.page:
            string = f'Page {self.page}; {self.per_page} of {self.total} results {string}'
        return string

    def first(self) -> Optional[Model]:
        """Returns the first result or None if there are no results"""
        return next(iter(self), None)

    def only(self) -> Optional[Model]:
        """Returns the single result that was found.

        :raises ValueError: If there are no results or more than one result
        """
        if len(self) != 1:
            raise ValueError('Expected exactly one result')
        return self.first()


class DAO(TransactionMixin):
    """A DAO implementation for SQLAlchemy to make your code less SQLly."""
    def __init__(self, model_class: type[Model], db: Session):
        self.model_class = model_class
        self.db = db
        if not hasattr(self.db, '_in_transaction'):
            self.db._in_transaction = False

    @property
    def query(self) -> Query[Any]:
        """Access the SQLAlchemy Query object for full SQLAlchemy functionality.

        :return: The Query for the current Session
        """
        return self.db.query(self.model_class)

    def _check_pk_arguments(self, pk_values: tuple) -> dict[str, Any]:
        """Validates that the number of primary key values matches the expected count.

        :param pk_values: The primary key values to validate (matching the order they are defined in the model)
        :return: A dictionary mapping primary key names to their values
        :raises InvalidArgumentCount: if the provided values do not align with the model's primary key
        """
        keys = self.model_class.get_pk_names()
        if len(pk_values) != len(keys):
            raise InvalidArgumentCount(len(keys), len(pk_values), f'{self.model_class.doc_name()} primary key')
        return {keys[i]: pk_values[i] for i in range(len(keys))}

    def create(self, *pk_values: Any) -> Model:
        """Creates a new entry for the given primary key.

        :param pk_values: Primary key values to represent the Model (in the order defined in the model)
        :return: The DAOModel entry that was newly added to the database
        :raises PrimaryKeyConflict: if an entry already exists for the primary key
        :raises InvalidArgumentCount: if the provided values do not align with the model's primary key
        """
        return self.create_with(**self._check_pk_arguments(pk_values))

    def create_with(self, insert: bool = True, **values: Any) -> Model:
        """Creates a new entry for the given primary key and property values.

        :param insert: False to avoid adding the model to the database
        :param values: The values to assign to the model
        :return: The new DAOModel
        :raises PrimaryKeyConflict: if an entry already exists for the primary key (does not apply if insert=False)
        """
        model = self.model_class(**retain_in_dict(values, *self.model_class.get_pk_names()))
        model.set_values(ignore_pk=True, **values)
        if insert:
            self.insert(model)
        return model

    def insert(self, model: Model) -> None:
        """Adds the given model to the database.

        :param model: The DAOModel entry to add
        :raises PrimaryKeyConflict: if an entry already exists for the primary key
        """
        if self.exists(model):
            raise PrimaryKeyConflict(model)
        self.db.add(model)
        if self._auto_commit:
            self.commit()
            self.db.refresh(model)

    def upsert(self, model: Model) -> None:
        """Updates the given model in the database or creates it if it does not exist.

        :param model: The DAOModel entry which may or may not exist
        """
        try:
            self.insert(model)
        except Conflict:
            self.commit()
        return model

    def rename(self, existing: Model, *new_pk_values: Any) -> None:
        """Updates the given model with new primary key values.

        :param existing: The model to rename
        :param new_pk_values: The new primary key values for the model
        :raises PrimaryKeyConflict: if an entry already exists for the new primary key
        """
        try:
            raise PrimaryKeyConflict(self.get(*new_pk_values))
        except NotFound:
            for k, v in zip(existing.get_pk_names(), new_pk_values):
                setattr(existing, k, v)
            self._commit_if_not_transaction()

    def exists(self, model: Model) -> bool:
        """Determines if a model exists in the database.

        :param model: The DAOModel entry in question
        :return: True if the model exists in the database, False otherwise
        """
        return bool(self.query.filter_by(**model.get_pk_dict()).count())

    def get(self, *pk_values: Any) -> Model:
        """Retrieves an entry from the database by its primary key.

        :param pk_values: The primary key values of the Model to fetch (in the order defined in the model)
        :return: The DAOModel entry that was retrieved
        :raises NotFound: if the model does not exist in the database
        :raises InvalidArgumentCount: if the provided values do not align with the model's primary key
        """
        return self.get_with(**self._check_pk_arguments(pk_values))

    def get_with(self, **values: Any) -> Model:
        """Retrieves an entry from the database and applies the given values to it.

        These changes are not committed to the database. Call commit() to do so.

        :param values: A dictionary containing the pk values of the requested model along with additional values to set
        :return: The DAOModel entry with the additional properties updated
        :raises NotFound: if the model does not exist in the database
        """
        pk = values_from_dict(*self.model_class.get_pk_names(), **values)
        model = self.query.get(pk)
        if model is None:
            raise NotFound(self.model_class(**values))
        model.set_values(ignore_pk=True, **values)
        return model

    def find(self,
             page: Optional[int] = None,
             per_page: Optional[int] = None,
             **filters: Any) -> SearchResults[Model]:
        """Searches all the DAOModel entries to return results.

        :param page: The number of the page to fetch
        :param per_page: How many results are on each page
        :param filters: Criteria to filter down the number of results
        :return: The SearchResults for the provided filters
        """
        query = self.query
        order = self.model_class.get_pk()
        foreign_tables = []

        # TODO: Add support for checking for specific values within foreign tables
        for key, value in filters.items():
            if key == 'order':  # TODO: rename to avoid collisions with actual column names
                order = self._order(value, foreign_tables)
            elif key == 'duplicate':
                query = self._count(query, value, foreign_tables, 'dupe').where(text(f'dupe.count > 1'))
            elif key == 'unique':
                query = self._count(query, value, foreign_tables, 'uniq').where(text(f'uniq.count <= 1'))
            else:  # TODO: Add logic for is_set and not_set that works for foreign values
                query = self._filter(query, key, value, foreign_tables)

        for table in dedupe(foreign_tables):
            query = query.join(table)

        query = query.order_by(*order)
        query = self.filter_find(query, **filters)

        total = query.count()
        if per_page:
            if not page or page < 1:
                page = 1
            query = query.offset((page - 1) * per_page).limit(per_page)
        elif page:
            raise MissingInput('Must specify how many results per page')

        return SearchResults(query.all(), total, page, per_page)

    def _order(self,
               value: str|Column|UnaryExpression|Iterable[str|Column|UnaryExpression],
               foreign_tables: list[DAOModel]) -> list[Column|UnaryExpression]:
        order = []
        if type(value) is str:
            value = value.split(', ')
        for column in ensure_iter(value):
            if type(column) is UnaryExpression:
                if self.model_class.find_searchable_column(column.element, foreign_tables) is not None:
                    order.append(column)
            else:
                order.append(self.model_class.find_searchable_column(column, foreign_tables))
        return order

    def _count(self, query: Query, prop: str, foreign_tables: list[DAOModel], alias: str) -> Query:
        column = self.model_class.find_searchable_column(prop, foreign_tables)
        subquery = (self.db.query(column, func.count(column).label('count'))
                    .group_by(column)
                    .subquery()
                    .alias(alias))
        return query.join(subquery, column == text(f'{alias}.{column.name}'))

    def _filter(self, query: Query, key: [str|Column], value: Any, foreign_tables: list[type[DAOModel]]) -> Query:
        column = self.model_class.find_searchable_column(key, foreign_tables)
        return query.filter(value.get_expression(column) if isinstance(value, ConditionOperator) else column == value)

    def filter_find(self, query: Query, **filters: Any) -> Query:
        """Overridable function to customize filtering.

        :param query: The session's SQLAlchemy Query
        :param filters: Any provided filter terms
        :return: The newly modified Query
        """
        return query

    def remove(self, model: Model) -> None:
        """Deletes the given model entry from the database.

        :param model: The DAOModel object to be deleted
        :raises NotFound: if the model does not exist in the database
        """
        if self.exists(model):
            self.db.delete(model)
        else:
            raise NotFound(model)
        self._commit_if_not_transaction()



from typing import Any, Iterable

import pytest
from sqlalchemy.testing.schema import Column

from daomodel import reference_of
from daomodel.util import names_of, values_from_dict, ensure_iter, dedupe, in_order, retain_in_dict, \
    remove_from_dict, mode
from tests.labeled_tests import labeled_tests
from tests.school_models import Person, Book


@pytest.mark.parametrize('column, expected', [
    (Person.name, 'person.name'),
    (Person.ssn, 'person.ssn'),
    (Book.owner, 'book.owner')
])
def test_reference_of(column: Column, expected: str):
    assert reference_of(column) == expected


@pytest.mark.parametrize('columns, expected', [
    ([], []),
    ([Column('one')], ['one']),
    ([Column('one'), Column('two'), Column('three')], ['one', 'two', 'three'])
])
def test_names_of(columns: list[Column], expected:  list[str]):
    assert names_of(columns) == expected


@pytest.mark.parametrize('elements, expected', [
    ([1, 2, 2, 3], 2),
    ([1, 1, 2, 2, 3], 1),
    (['a', 'b', 'b', 'c', 'c', 'c'], 'c'),
    ([True, True, False, True], True),
    ([1], 1)
])
def test_mode(elements: list, expected: Any):
    assert mode(elements) == expected


@pytest.mark.parametrize('keys, expected', [
    ((), ()),
    (('b',), (2,)),
    (('a', 'c'), (1, 3)),
    (('b', 'c', 'a'), (2, 3, 1))
])
def test_values_from_dict(keys: tuple[str, ...], expected: tuple):
    assert values_from_dict(*keys, a=1, b=2, c=3) == expected


@pytest.mark.parametrize('keys, expected', [
    ((), {}),
    (('b',), {'b': 2}),
    (('a', 'c'), {'a':1, 'c':3}),
    (('b', 'c', 'a'), {'a':1, 'b':2, 'c':3})
])
def test_retain_in_dict(keys: tuple[str, ...], expected: tuple):
    assert retain_in_dict({'a': 1, 'b': 2, 'c': 3}, *keys) == expected


@pytest.mark.parametrize('keys, expected', [
    ((), {'a':1, 'b':2, 'c':3}),
    (('b',), {'a':1, 'c':3}),
    (('a', 'c'), {'b': 2}),
    (('b', 'c', 'a'), {})
])
def test_remove_from_dict(keys: tuple[str, ...], expected: tuple):
    assert remove_from_dict({'a': 1, 'b': 2, 'c': 3}, *keys) == expected


@pytest.mark.parametrize('elements, expected', [
    ([1, 2, 3], [1, 2, 3]),
    ({1, 2, 3}, {1, 2, 3}),
    ((1, 2, 3), (1, 2, 3)),
    ([], []),
    ({}, {}),
    ((), ()),
    (1, [1]),
    (None, [None]),
    ('element', ['element'])
])
def test_ensure_iter(elements: Any, expected: Iterable[Any]):
    assert ensure_iter(elements) == expected


@pytest.mark.parametrize('elements, expected', [
    ([1, 2, 3], [1, 2, 3]),
    ([1, 1, 2, 2, 3, 3, 3], [1, 2, 3]),
    (['one', 'two', 'two', 'three'], ['one', 'two', 'three']),
    (['one', 1, 'one', 'two', 2, 'three', 2, 'three'], ['one', 1, 'two', 2, 'three']),
    ([], []),
])
def test_dedupe(elements: list, expected: list):
    assert dedupe(elements) == expected

@labeled_tests({
    'empty': [
        (set(), [], []),
        (set(), [1], []),
        ({1}, [], []),
    ],
    'single item': [
        ({1}, [1], [1]),
        ({'a'}, ['a'], ['a'])
    ],
    'multiple items': [
        ({1, 2, 3}, [1, 2, 3], [1, 2, 3]),
        ({1, 2, 3}, [3, 2, 1], [3, 2, 1]),
        ({'a', 'b', 'c'}, ['a', 'b', 'c'], ['a', 'b', 'c']),
        ({'a', 'b', 'c'}, ['c', 'b', 'a'], ['c', 'b', 'a'])
    ],
    'repeated items': [
        ([1, 1, 2, 2, 3, 3], [1, 2, 3], [1, 2, 3]),
        ([1, 1, 2, 2, 3, 3], [3, 2, 1], [3, 2, 1]),
        (['a', 'b', 'c', 'b', 'a'], ['a', 'b', 'c'], ['a', 'b', 'c']),
        (['a', 'b', 'c', 'b', 'a'], ['c', 'b', 'a'], ['c', 'b', 'a']),
        ([True, True, False, True, False, False], [False, True], [False, True])
    ],
    'items missing from order': [
        ({1, 2, 3, 4}, [1, 2, 3], [1, 2, 3]),
        ({1, 2, 3, 4}, [3, 2, 1], [3, 2, 1])
    ],
    'extraneous items in order': [
        ({1, 3}, [1, 2, 3], [1, 3]),
        ({1, 3}, [3, 2, 1], [3, 1])
    ],
    'mixed scenarios': [
        ({4, 1, 1, 3, 4}, [1, 2, 3], [1, 3]),
        ({1, 3, 1, 4}, [3, 2, 1], [3, 1])
    ]
})
def test_in_order(items: Iterable, order: list, expected: list):
    assert in_order(items, order) == expected



from typing import Iterable, Any, Optional

import pytest
from sqlalchemy import Column
from sqlmodel import Field

from daomodel import DAOModel, names_of, UnsearchableError, reference_of, ColumnBreadcrumbs
from daomodel.fields import Identifier, Unsearchable
from tests.labeled_tests import labeled_tests, Expected


class Model(DAOModel):
    pass


class SimpleModel(DAOModel, table=True):
    pkA: Identifier[int]

simple_instance = SimpleModel(pkA=23)


class ForeignKEYModel(DAOModel, table=True):
    pkB: Identifier[int]
    prop: str
    sensitive_prop: Unsearchable[str]
    fk: SimpleModel


class BaseModel(DAOModel):
    prop1: str


class ComplicatedModel(BaseModel, table=True):
    pkC: Identifier[int]
    pkD: Identifier[int]
    prop2: Unsearchable[Optional[str]]
    fkA: int = Field(foreign_key='simple_model.pkA')
    fkB: int = Field(foreign_key='foreign_key_model.pkB')

    class Meta:
        searchable_relations = {
            ForeignKEYModel.prop,
            ForeignKEYModel.sensitive_prop,
            (ForeignKEYModel, SimpleModel.pkA)
        }

complicated_instance = ComplicatedModel(pkC=17, pkD=76, prop1='prop', prop2='erty', fkA=23, fkB=32)


class MultiForeignKEYModel(DAOModel, table=True):
    fkC: int = Field(primary_key=True, foreign_key='complicated_model.pkC')
    fkD: int = Field(primary_key=True, foreign_key='complicated_model.pkD')
    fk_prop: str = Field(foreign_key='foreign_key_model.prop')


def test_tablename():
    assert SimpleModel.__tablename__ == 'simple_model'
    assert ForeignKEYModel.__tablename__ == 'foreign_key_model'


@labeled_tests({
    'true': [
        (SimpleModel, SimpleModel.pkA, True),
        (ForeignKEYModel, ForeignKEYModel.pkB, True),
        (ForeignKEYModel, ForeignKEYModel.prop, True),
        (ForeignKEYModel, ForeignKEYModel.fk, True),
        (ComplicatedModel, ComplicatedModel.pkC, True),
        (ComplicatedModel, ComplicatedModel.pkD, True),
        (ComplicatedModel, ComplicatedModel.prop1, True),
        (ComplicatedModel, ComplicatedModel.prop2, True),
        (ComplicatedModel, ComplicatedModel.fkA, True),
        (ComplicatedModel, ComplicatedModel.fkB, True),
        (MultiForeignKEYModel, MultiForeignKEYModel.fkC, True),
        (MultiForeignKEYModel, MultiForeignKEYModel.fkD, True),
        (MultiForeignKEYModel, MultiForeignKEYModel.fk_prop, True)
    ],
    'false': [
        (SimpleModel, ForeignKEYModel.pkB, False),
        (SimpleModel, ForeignKEYModel.prop, False),
        (SimpleModel, ComplicatedModel.pkC, False),
        (SimpleModel, ComplicatedModel.pkD, False),
        (SimpleModel, ComplicatedModel.prop1, False),
        (SimpleModel, ComplicatedModel.prop2, False),
        (SimpleModel, ComplicatedModel.fkB, False),
        (SimpleModel, MultiForeignKEYModel.fk_prop, False),
        (ForeignKEYModel, ComplicatedModel.pkC, False),
        (ForeignKEYModel, ComplicatedModel.pkD, False),
        (ForeignKEYModel, ComplicatedModel.prop1, False),
        (ForeignKEYModel, ComplicatedModel.prop2, False),
        (ForeignKEYModel, ComplicatedModel.fkA, False),
        (ForeignKEYModel, MultiForeignKEYModel.fkC, False),
        (ForeignKEYModel, MultiForeignKEYModel.fkD, False),
        (ComplicatedModel, ForeignKEYModel.prop, False),
        (ComplicatedModel, MultiForeignKEYModel.fk_prop, False),
        (MultiForeignKEYModel, SimpleModel.pkA, False),
        (MultiForeignKEYModel, ComplicatedModel.prop1, False),
        (MultiForeignKEYModel, ComplicatedModel.prop2, False),
        (MultiForeignKEYModel, ComplicatedModel.fkA, False),
        (MultiForeignKEYModel, ComplicatedModel.fkB, False)
    ],
    'foreign keys': [
        (ForeignKEYModel, SimpleModel.pkA, False),
        (ComplicatedModel, SimpleModel.pkA, False),
        (ComplicatedModel, ForeignKEYModel.pkB, False),
        (MultiForeignKEYModel, ComplicatedModel.pkC, False),
        (MultiForeignKEYModel, ComplicatedModel.pkD, False),
        (MultiForeignKEYModel, ForeignKEYModel.prop, False)
    ],
    'foreign references': [
        (SimpleModel, ForeignKEYModel.fk, False),
        (ForeignKEYModel, ComplicatedModel.fkB, False),
        (ForeignKEYModel, MultiForeignKEYModel.fk_prop, False),
        (ComplicatedModel, MultiForeignKEYModel.fkC, False),
        (ComplicatedModel, MultiForeignKEYModel.fkD, False)
    ],
    'inherited': [
        (ComplicatedModel, ComplicatedModel.prop1, True),
        (BaseModel, ComplicatedModel.prop1, False),
        (BaseModel, ComplicatedModel.prop2, False)
    ]
})
def test_has_column(model: DAOModel, column: Column, expected: bool):
    assert model.has_column(column) is expected


@labeled_tests({
    'single word':
        (Model, 'model'),
    'multiple words': [
        (SimpleModel, 'simple_model'),
        (ComplicatedModel, 'complicated_model')
    ],
    'acronym': [
        (ForeignKEYModel, 'foreign_key_model'),
        (MultiForeignKEYModel, 'multi_foreign_key_model')
    ]
})
def test_normalized_name(model: type[DAOModel], expected: str):
    assert model.normalized_name() == expected


@labeled_tests({
    'single word':
        (Model, 'Model'),
    'multiple words': [
        (SimpleModel, 'Simple Model'),
        (ComplicatedModel, 'Complicated Model')
    ],
    'acronym': [
        (ForeignKEYModel, 'Foreign Key Model'),
        (MultiForeignKEYModel, 'Multi Foreign Key Model')
    ]
})
def test_doc_name(model: type[DAOModel], expected: str):
    assert model.doc_name() == expected


@labeled_tests({
    'single column':
        (SimpleModel, ['pkA']),
    'multiple columns':
        (ComplicatedModel, ['pkC', 'pkD'])
})
def test_get_pk_names__single_column(model: type[DAOModel], expected: list[str]):
    assert model.get_pk_names() == expected


@labeled_tests({
    'single column':
        (simple_instance, (23,)),
    'multiple columns':
        (complicated_instance, (17, 76))
})
def test_get_pk_values(model: DAOModel, expected: tuple[int, ...]):
    assert model.get_pk_values() == expected


@labeled_tests({
    'single column':
        (simple_instance, {'pkA': 23}),
    'multiple columns':
        (complicated_instance, {'pkC': 17, 'pkD': 76})
})
def test_get_pk_dict(model: DAOModel, expected: dict[str, int]):
    assert model.get_pk_dict() == expected


@labeled_tests({
    'single column':
        (ForeignKEYModel, {'pkA'}),
    'multiple columns':
        (ComplicatedModel, {'pkA', 'pkB'})
})
def test_get_fks(model: type[DAOModel], expected: set[str]):
    assert set(names_of(model.get_fks())) == expected


@labeled_tests({
    'single column':
        (ForeignKEYModel, {'fk'}),
    'multiple columns':
        (ComplicatedModel, {'fkA', 'fkB'})
})
def test_get_fk_properties(model: type[DAOModel], expected: set[str]):
    assert set(names_of(model.get_fk_properties())) == expected


def to_str(columns: Iterable[Column]):
    """Converts columns to strings instead to make assert comparisons simpler"""
    return {reference_of(column) for column in columns}


@labeled_tests({
    'single column':
        (ForeignKEYModel, SimpleModel, {ForeignKEYModel.fk}),
    'some applicable columns': [
        (ComplicatedModel, SimpleModel, {ComplicatedModel.fkA}),
        (ComplicatedModel, ForeignKEYModel, {ComplicatedModel.fkB})
    ],
    'no applicable columns':
        (MultiForeignKEYModel, SimpleModel, {}),
    'non primary key':
        (MultiForeignKEYModel, ForeignKEYModel, {MultiForeignKEYModel.fk_prop}),
    'multiple columns':
        (MultiForeignKEYModel, ComplicatedModel, {MultiForeignKEYModel.fkC, MultiForeignKEYModel.fkD})
})
def test_get_references_of(model: type[DAOModel], reference: type[DAOModel], expected: set[Column]):
    assert to_str(model.get_references_of(reference)) == to_str(expected)


@labeled_tests({
    'single column':
        (SimpleModel, ['pkA']),
    'multiple columns':
        (ForeignKEYModel, ['pkB', 'prop', 'sensitive_prop', 'fk']),
    'inherited columns':
        (ComplicatedModel, ['prop1', 'pkC', 'pkD', 'prop2', 'fkA', 'fkB'])
})
def test_get_properties(model: type[DAOModel], expected: list[str]):
    assert names_of(model.get_properties()) == expected


@labeled_tests({
    'primary key': [
        (simple_instance, SimpleModel.pkA, 23),
        (complicated_instance, ComplicatedModel.pkC, 17),
        (complicated_instance, ComplicatedModel.pkD, 76)
    ],
    'foreign key': [
        (complicated_instance, ComplicatedModel.fkA, 23),
        (complicated_instance, ComplicatedModel.fkB, 32)
    ],
    'inherited column':
        (complicated_instance, ComplicatedModel.prop1, 'prop'),
    'standard column':
        (complicated_instance, ComplicatedModel.prop2, 'erty'),
    'by str':
        (complicated_instance, 'prop2', 'erty')
})
def test_get_value_of(model: DAOModel, column: Column, expected: Any):
    assert model.get_value_of(column) == expected


@labeled_tests({
    'no columns': [
        (simple_instance, [], {}),
        (complicated_instance, [], {})
    ],
    'single column': [
        (simple_instance, ['pkA'], {'pkA': 23}),
        (complicated_instance, ['pkC'], {'pkC': 17}),
        (complicated_instance, ['pkD'], {'pkD': 76}),
        (complicated_instance, ['prop1'], {'prop1': 'prop'}),
        (complicated_instance, ['prop2'], {'prop2': 'erty'}),
        (complicated_instance, ['fkA'], {'fkA': 23}),
        (complicated_instance, ['fkB'], {'fkB': 32}),
    ],
    'multiple columns':
        (complicated_instance, ['pkC', 'pkD', 'prop1', 'prop2', 'fkA', 'fkB'], {
            'pkC': 17,
            'pkD': 76,
            'prop1': 'prop',
            'prop2': 'erty',
            'fkA': 23,
            'fkB': 32
        })
})
def test_get_values_of(model: DAOModel, columns: list[Column], expected: Any):
    assert model.get_values_of(columns) == expected


@labeled_tests({
    'no diff': [
        (complicated_instance, complicated_instance, {}),
        (complicated_instance, ComplicatedModel(pkC=17, pkD=76, prop1='prop', prop2='erty', fkA=23, fkB=32), {}),
        (complicated_instance, ComplicatedModel(pkC=18, pkD=77, prop1='prop', prop2='erty', fkA=23, fkB=32), {})
    ],
    'single column': [
        (
                complicated_instance,
                ComplicatedModel(pkC=17, pkD=76, prop1='new', prop2='erty', fkA=23, fkB=32),
                {'prop1': ('prop', 'new')}
        ), (
                complicated_instance,
                ComplicatedModel(pkC=17, pkD=76, prop1='prop', prop2='new', fkA=23, fkB=32),
                {'prop2': ('erty', 'new')}
        ), (
                complicated_instance,
                ComplicatedModel(pkC=17, pkD=76, prop1='prop', prop2='erty', fkA=24, fkB=32),
                {'fkA': (23, 24)}
        ), (
                complicated_instance,
                ComplicatedModel(pkC=17, pkD=76, prop1='prop', prop2='erty', fkA=23, fkB=33),
                {'fkB': (32, 33)}
        )
    ],
    'multiple columns': [
        (
                complicated_instance,
                ComplicatedModel(pkC=17, pkD=76, prop1='new', prop2='prop', fkA=23, fkB=32),
                {'prop1': ('prop', 'new'), 'prop2': ('erty', 'prop')}
        ), (
                complicated_instance,
                ComplicatedModel(pkC=17, pkD=76, prop1='prop', prop2='erty', fkA=0, fkB=0),
                {'fkA': (23, 0), 'fkB': (32, 0)}
        )
    ],
    'none values': [
        (
                complicated_instance,
                ComplicatedModel(pkC=17, pkD=76, prop1='new', prop2=None, fkA=23, fkB=32),
                {'prop1': ('prop', 'new'), 'prop2': ('erty', None)}
        ), (
                ComplicatedModel(pkC=17, pkD=76, prop1='new', prop2=None, fkA=23, fkB=32),
                complicated_instance,
                {'prop1': ('new', 'prop'), 'prop2': (None, 'erty')}
        )
    ],
    'unset values': [
        (
                complicated_instance,
                ComplicatedModel(pkC=17, pkD=76, prop1='new', fkA=23, fkB=32),
                {'prop1': ('prop', 'new'), 'prop2': ('erty', None)}
        ), (
                ComplicatedModel(pkC=17, pkD=76, prop1='new', fkA=23, fkB=32),
                complicated_instance,
                {'prop1': ('new', 'prop'), 'prop2': (None, 'erty')}
        )
    ]
})
def test_compare(model: DAOModel, other: DAOModel, expected: dict[str, tuple[Any, Any]]):
    assert model.compare(other) == expected


@labeled_tests({
    'no diff': [
        (simple_instance, simple_instance, {}),
        (complicated_instance, complicated_instance, {}),
        (complicated_instance, ComplicatedModel(pkC=17, pkD=76, prop1='prop', prop2='erty', fkA=23, fkB=32), {})
    ],
    'single column': [
        (
                simple_instance,
                SimpleModel(pkA=24),
                {'pkA': (23, 24)}
        ), (
                complicated_instance,
                ComplicatedModel(pkC=18, pkD=76, prop1='prop', prop2='erty', fkA=23, fkB=32),
                {'pkC': (17, 18)}
        ), (
                complicated_instance,
                ComplicatedModel(pkC=17, pkD=77, prop1='prop', prop2='erty', fkA=23, fkB=32),
                {'pkD': (76, 77)}
        )
    ],
    'multiple columns': [
        (
                complicated_instance,
                ComplicatedModel(pkC=18, pkD=86, prop1='prop', prop2='erty', fkA=23, fkB=32),
                {'pkC': (17, 18), 'pkD': (76, 86)}
        )
    ]
})
def test_compare__include_pk(model: DAOModel, other: DAOModel, expected: dict[str, tuple[Any, Any]]):
    assert model.compare(other, include_pk=True) == expected


@labeled_tests({
    'single column':
        (SimpleModel, ['pkA']),
    'multiple columns':
        (ForeignKEYModel, ['pkB', 'prop', 'fk'])
})
def test_get_searchable_properties(model: type[DAOModel], expected: list[str]):
    assert names_of(model.get_searchable_properties()) == expected


@labeled_tests({
    'mapped relation': ForeignKEYModel.prop,
    'unsearchable mapped relation': ForeignKEYModel.sensitive_prop,
    'breadcrumbs': (ForeignKEYModel, SimpleModel.pkA)
})
def test_get_searchable_properties__meta(expected: Column | ColumnBreadcrumbs):
    assert expected in ComplicatedModel.get_searchable_properties()


@labeled_tests({
    'column': [ Expected([]),
        ComplicatedModel.pkC,
        ComplicatedModel.pkD,
        ComplicatedModel.prop1,
        ComplicatedModel.fkA,
        ComplicatedModel.fkB
    ],
    'column reference': [ Expected([]),
        'complicated_model.pkC',
        'complicated_model.pkD',
        'complicated_model.prop1',
        'complicated_model.fkA',
        'complicated_model.fkB'
    ],
    'column name': [ Expected([]),
        'pkC',
        'pkD',
        'prop1',
        'fkA',
        'fkB'
    ],
    'foreign property': [ Expected([ForeignKEYModel.normalized_name()]),
        ForeignKEYModel.prop,
        'foreign_key_model.prop'
    ],
    'nested foreign property': [ Expected([ForeignKEYModel.normalized_name(), SimpleModel.normalized_name()]),
        SimpleModel.pkA,
        'simple_model.pkA'
    ]
})
def test_find_searchable_column(prop: str|Column, expected: list[str]):
    foreign_tables = []
    assert ComplicatedModel.find_searchable_column(prop, foreign_tables) is not None
    assert [t.name for t in foreign_tables] == expected


def test_find_searchable_column__foreign_without_table():
    with pytest.raises(UnsearchableError):
        ComplicatedModel.find_searchable_column('prop', [])


@labeled_tests({
    'single column':
        (SimpleModel, (23,), {'pkA': 23}),
    'multiple columns':
        (ComplicatedModel, (17, 76), {'pkC': 17, 'pkD': 76})
})
def test_pk_values_to_dict__single_column(model: type[DAOModel], args: tuple[int, ...], expected: dict[str, int]):
    assert model.pk_values_to_dict(args) == expected


def test_copy_model():
    other = ComplicatedModel(pkC=12, pkD=34, prop1='different', prop2='values', fkA=1, fkB=2)
    other.copy_model(complicated_instance)
    assert other.pkC == 12
    assert other.pkD == 34
    assert other.prop1 == 'prop'
    assert other.prop2 == 'erty'
    assert other.fkA == 23
    assert other.fkB == 32
    assert complicated_instance.pkC == 17
    assert complicated_instance.pkD == 76
    assert complicated_instance.prop1 == 'prop'
    assert complicated_instance.prop2 == 'erty'
    assert complicated_instance.fkA == 23
    assert complicated_instance.fkB == 32


def test_copy_model__specific_fields():
    other = ComplicatedModel(pkC=12, pkD=34, prop1='different', prop2='values', fkA=1, fkB=2)
    other.copy_model(complicated_instance, 'prop1', 'fkA')
    assert other.pkC == 12
    assert other.pkD == 34
    assert other.prop1 == 'prop'
    assert other.prop2 == 'values'
    assert other.fkA == 23
    assert other.fkB == 2


def test_copy_model__pk():
    other = ComplicatedModel(pkC=12, pkD=34, prop1='different', prop2='values', fkA=1, fkB=2)
    other.copy_model(complicated_instance, 'pkD')
    assert other.pkC == 12
    assert other.pkD == 76
    assert other.prop1 == 'different'
    assert other.prop2 == 'values'
    assert other.fkA == 1
    assert other.fkB == 2


def test_set_values():
    other = ComplicatedModel(pkC=12, pkD=34, prop1='different', prop2='values', fkA=1, fkB=2)
    other.set_values(prop1='new', fkB=3)
    assert other.model_dump() == {
        'pkC': 12,
        'pkD': 34,
        'prop1': 'new',
        'prop2': 'values',
        'fkA': 1,
        'fkB': 3
    }


def test_set_values__extra_values():
    other = ComplicatedModel(pkC=12, pkD=34, prop1='different', prop2='values', fkA=1, fkB=2)
    other.set_values(prop1='new', other='extra')
    assert other.model_dump() == {
        'pkC': 12,
        'pkD': 34,
        'prop1': 'new',
        'prop2': 'values',
        'fkA': 1,
        'fkB': 2
    }


def test_set_values__ignore_pk():
    other = ComplicatedModel(pkC=12, pkD=34, prop1='different', prop2='values', fkA=1, fkB=2)
    other.set_values(ignore_pk=True, pkC=0, prop1='new')
    assert other.model_dump() == {
        'pkC': 12,
        'pkD': 34,
        'prop1': 'new',
        'prop2': 'values',
        'fkA': 1,
        'fkB': 2
    }


def test_set_values__pk():
    other = ComplicatedModel(pkC=12, pkD=34, prop1='different', prop2='values', fkA=1, fkB=2)
    other.set_values(pkC=0)
    assert other.model_dump() == {
        'pkC': 0,
        'pkD': 34,
        'prop1': 'different',
        'prop2': 'values',
        'fkA': 1,
        'fkB': 2
    }


@labeled_tests({
    'single column': (
            simple_instance,
            SimpleModel(pkA=23),
            (
                    SimpleModel(pkA=32),
                    SimpleModel(pkA=45)
            )
    ),
    'multiple columns': (
            complicated_instance,
            ComplicatedModel(pkC=17, pkD=76, prop1='different', prop2='values', fkA=1, fkB=2),
            (
                    ComplicatedModel(pkC=17, pkD=89, prop1='prop', prop2='erty', fkA=23, fkB=32),
                    ComplicatedModel(pkC=17, pkD=89, prop1='prop', prop2='erty', fkA=23, fkB=32)
            )
    )
})
def test___eq____hash__(model: DAOModel, equivalent: DAOModel, different: tuple[DAOModel, ...]):
    assert model == equivalent
    assert hash(model) == hash(equivalent)
    for instance in different:
        assert model != instance
        assert hash(model) != hash(instance)


@labeled_tests({
    'single column':
        (simple_instance, '23'),
    'multiple columns':
        (complicated_instance, '(17, 76)')
})
def test___str___single_column(model: DAOModel, expected: str):
    assert str(model) == expected



import warnings
from typing import Iterable, Any, OrderedDict

from sqlalchemy import Column, ColumnElement
from sqlmodel import or_, and_


class MissingInput(Exception):
    """Indicates that required information was not provided."""
    def __init__(self, detail: str):
        self.detail = detail


class InvalidArgumentCount(Exception):
    """Indicates that an incorrect number of arguments was provided."""
    def __init__(self, expected: int, got: int, context: str = None):
        self.detail = f'Expected {expected} values, got {got}'
        if context:
            self.detail += f' for {context}'


class UnsupportedFeatureError(Exception):
    """Indicates that a feature is not yet supported.

    If you think the feature should be implemented, please open an issue on GitHub.
    """
    def __init__(self, detail: str):
        self.detail = detail + (' Note: This functionality is not yet supported. '
                                'Please submit a request through GitHub if you would like it implemented.')


def reference_of(column: Column) -> str:
    """
    Prepares a str reference of a column.

    :param column: The column to convert
    :return: The 'table.column' notation of the Column
    """
    return f'{column.table.name}.{column.name}'


def names_of(properties: Iterable[Column]) -> list[str]:
    """
    Reduces Columns to just their names.

    :param properties: A group of Columns
    :return: A list of names matching the order of the Columns provided
    """
    return [p.name for p in properties]


def mode(values: Iterable[Any]) -> Any:
    """Determines the most frequently occurring value within the provided iterable.

    In the case of a tie (multiple values with the same highest frequency),
    the function returns the first value encountered with that frequency.

    :param values: An iterable containing values to evaluate
    :return: The most common value from the provided iterable
    """
    values = list(values)
    return max(values, key=values.count)


def values_from_dict(*keys: Any, **values: Any) -> tuple:
    """Pulls specific values from a dictionary.

    :param keys: The keys to read from the dict
    :param values: The dictionary containing the values
    :return: A tuple of values read from the dict, in the same order as keys
    """
    result = []
    for key in keys:
        if key in values:
            result.append(values[key])
        else:
            raise MissingInput(f'Requested key {key} not found in dictionary')
    return tuple(result)


def retain_in_dict(d: dict[Any, Any], *keys: Any) -> dict[Any, Any]:
    """Filters a dictionary to specified keys.

    The source dict remains unmodified.

    :param d: The dictionary to filter down
    :param keys: The target keys for the new dict
    :return: The reduced values as a new dict
    """
    return {key: d[key] for key in keys if key in d}


def remove_from_dict(d: dict[Any, Any], *keys: Any) -> dict[Any, Any]:
    """Removes specified key/value pairs from a dictionary.

    The source dict remains unmodified.

    :param d: The dictionary to adjust
    :param keys: The keys to remove from the dict
    :return: The modified values as a new dict
    """
    return {k: v for k, v in d.items() if k not in keys}


def ensure_iter(elements: Any):
    """Ensures that the provided argument is iterable.

    Single, non-Iterable items are converted to a single-item list.
    In this context, a str is not considered to be Iterable.

    :param elements: The input that may or may not be Iterable
    :return: The provided Iterable or a single item list
    """
    if not isinstance(elements, Iterable) or type(elements) is str:
        elements = [elements]
    return elements


def dedupe(original: list) -> list:
    """Creates a filtered copy of a list that does not include duplicates.

    :param original: The list to filter
    :return: a new list that maintains order but is guaranteed to have no duplicates
    """
    return list(OrderedDict.fromkeys(original))


def in_order(original: Iterable, order: list) -> list:
    """Returns provided items as an ordered list.

    Repeated items will be deduplicated.
    Items not defined within the order will be excluded.
    The order is allowed to contain extraneous items that aren't applicable to the provided items.

    :param original: The (likely unordered) collection of items
    :param order: The defined order of items
    :return: a new list of the items following the defined order
    """
    return [item for item in order if item in original]


def next_id() -> None:
    """Indicates to the model that an id should be auto-incremented"""
    return None


class ConditionOperator:
    """A utility class to easily generate common expressions"""
    def __init__(self, *values: Any):
        self.values = values

    def get_expression(self, column: ColumnElement) -> ColumnElement:
        """Builds and returns the appropriate expression.

        :param column: The column on which to evaluate
        :return: the expression
        """
        raise NotImplementedError('Must implement `get_expression` in subclass')


class GreaterThan(ConditionOperator):
    def get_expression(self, column: ColumnElement) -> ColumnElement:
        return column > self.values[0]


class GreaterThanEqualTo(ConditionOperator):
    def get_expression(self, column: ColumnElement) -> ColumnElement:
        return column >= self.values[0]


class LessThan(ConditionOperator):
    def get_expression(self, column: ColumnElement) -> ColumnElement:
        return column < self.values[0]


class LessThanEqualTo(ConditionOperator):
    def get_expression(self, column: ColumnElement) -> ColumnElement:
        return column <= self.values[0]


class Between(ConditionOperator):
    def get_expression(self, column: ColumnElement) -> ColumnElement:
        lower_bound, upper_bound = self.values
        return and_(column >= lower_bound, column <= upper_bound)


class AnyOf(ConditionOperator):
    def get_expression(self, column: ColumnElement) -> ColumnElement:
        return or_(*[column == value for value in self.values])


class NoneOf(ConditionOperator):
    def get_expression(self, column: ColumnElement) -> ColumnElement:
        return and_(*[column != value for value in self.values])


class IsSet(ConditionOperator):
    """Expression to filter to rows that have a value set for a specific Column"""
    def get_expression(self, column: ColumnElement) -> ColumnElement:
        return or_(column == True, and_(column != None, column != False))


class NotSet(ConditionOperator):
    """Expression to filter to rows that have no value set for a specific Column"""
    def get_expression(self, column: ColumnElement) -> ColumnElement:
        return or_(column == False, column == None)


is_set = IsSet()
not_set = NotSet()



import pytest
from sqlalchemy.exc import InvalidRequestError

from daomodel.dao import DAO, NotFound, Conflict, PrimaryKeyConflict
from daomodel.util import MissingInput, InvalidArgumentCount, next_id
from tests.conftest import TestDAOFactory
from tests.school_models import Student, Person, Book


def test_create__single_column(daos: TestDAOFactory):
    dao = daos[Student]
    model = dao.create(100)
    assert model == Student(id=100)
    daos.assert_in_db(Student, 100)


def test_create__multi_column(daos: TestDAOFactory):
    model = daos[Person].create('John', 23)
    assert model == Person(name='John', age=23)
    daos.assert_in_db(Person, 'John', 23)


def test_create__default_fields(daos: TestDAOFactory):
    model = daos[Student].create(100)
    assert model.active
    daos.assert_in_db(Student, 100, active=True)


def test_create__required_fields(daos: TestDAOFactory):
    with pytest.raises(InvalidArgumentCount) as error:
        daos[Person].create('John')
    assert 'Expected 2 values, got 1' in str(error.value.detail)


def test_create__too_many_args(daos: TestDAOFactory):
    with pytest.raises(InvalidArgumentCount) as error:
        daos[Student].create(100, 'extra', 'args')
    assert 'Expected 1 values, got 3' in str(error.value.detail)


def test_create_with(daos: TestDAOFactory):
    model = daos[Student].create_with(id=100, name='Bob', active=False)
    assert model == Student(id=100, name='Bob', active=False)
    daos.assert_in_db(Student, 100, name='Bob', active=False)


def test_create_with__no_insert(daos: TestDAOFactory):
    model = daos[Student].create_with(False, id=100)
    assert model == Student(id=100)
    daos.assert_not_in_db(Student, 100)


def test_create_with__no_commit(daos: TestDAOFactory):
    dao = daos[Student]
    dao.start_transaction()
    model = dao.create_with(id=100)
    assert model == Student(id=100)
    daos.assert_not_in_db(Student, 100)


def test_insert(daos: TestDAOFactory):
    daos[Student].insert(Student(id=100))
    daos.assert_in_db(Student, 100)


def test_insert__default_fields(daos: TestDAOFactory):
    model = Student(id=100)
    daos[Student].insert(model)
    assert model.active
    daos.assert_in_db(Student, 100, active=True)


def test_insert__conflict(daos: TestDAOFactory):
    dao = daos[Student]
    dao.insert(Student(id=100))
    daos.assert_in_db(Student, 100)
    with pytest.raises(PrimaryKeyConflict):
        dao.insert(Student(id=100))


def test_insert__unique_constraint_violation(daos: TestDAOFactory):
    dao = daos[Person]
    dao.create_with(name='John', age=23, ssn='123-45-6789')
    with pytest.raises(Conflict):
        dao.create_with(name='Bob', age=32, ssn='123-45-6789')


def test_upsert__new(daos: TestDAOFactory):
    daos[Student].upsert(Student(id=100, name='Bob'))
    daos.assert_in_db(Student, 100, name='Bob')


def test_upsert__existing(daos: TestDAOFactory):
    dao = daos[Student]
    model = dao.create(100)
    model.name = 'Bob'
    daos.assert_in_db(Student, 100, name=None)
    dao.upsert(model)
    daos.assert_in_db(Student, 100, name='Bob')


def test_rename(daos: TestDAOFactory):
    dao = daos[Student]
    model = dao.create(100)
    daos.assert_in_db(Student, 100)
    daos.assert_not_in_db(Student, 200)
    dao.rename(model, 200)
    daos.assert_not_in_db(Student, 100)
    daos.assert_in_db(Student, 200)


def test_rename__multiple_columns(daos: TestDAOFactory):
    dao = daos[Person]
    model = dao.create('Alfred', 51)
    daos.assert_in_db(Person, 'Alfred', 51)
    daos.assert_not_in_db(Person, 'Al', 51)

    dao.rename(model, 'Al', 51)
    daos.assert_not_in_db(Person, 'Alfred', 51)
    daos.assert_in_db(Person, 'Al', 51)

    dao.rename(model, 'Fred', 52)
    daos.assert_not_in_db(Person, 'Al', 51)
    daos.assert_in_db(Person, 'Fred', 52)

    dao.rename(model, 'Fred', 53)
    daos.assert_not_in_db(Person, 'Fred', 52)
    daos.assert_in_db(Person, 'Fred', 53)


def test_rename__keep_property_values(daos: TestDAOFactory):
    dao = daos[Student]
    model = dao.create_with(id=100, name='Bob', gender='m', active=False)
    daos.assert_in_db(Student, 100, name='Bob', gender='m', active=False)
    daos.assert_not_in_db(Student, 200)
    dao.rename(model, 200)
    daos.assert_not_in_db(Student, 100)
    daos.assert_in_db(Student, 200, name='Bob', gender='m', active=False)


def test_rename__cascade_foreign_reference(daos: TestDAOFactory):
    student_dao = daos[Student]
    model = student_dao.create(100)
    daos[Book].create_with(name='Algebra I', subject='Math', owner=100)
    daos.assert_in_db(Book, 'Algebra I', subject='Math', owner=100)
    student_dao.rename(model, 200)
    daos.assert_in_db(Book, 'Algebra I', subject='Math', owner=200)


def test_rename__keep_foreign_key(daos: TestDAOFactory):
    daos[Student].create(100)
    book_dao = daos[Book]
    book = book_dao.create_with(name='Algebra I', subject='Math', owner=100)
    daos.assert_in_db(Book, 'Algebra I', subject='Math', owner=100)
    daos.assert_not_in_db(Book, 'Algebra II')
    book_dao.rename(book, 'Algebra II')
    daos.assert_not_in_db(Book, 'Algebra I')
    daos.assert_in_db(Book, 'Algebra II', subject='Math', owner=100)


def test_rename__already_exists(daos: TestDAOFactory):
    dao = daos[Student]
    model = dao.create(100)
    dao.create(200)
    daos.assert_in_db(Student, 100)
    daos.assert_in_db(Student, 200)
    with pytest.raises(PrimaryKeyConflict):
        dao.rename(model, 200)


def test_exists__true(daos: TestDAOFactory):
    dao = daos[Student]
    model = dao.create(100)
    assert dao.exists(model)


def test_exists__false(daos: TestDAOFactory):
    dao = daos[Student]
    model = dao.create_with(id=100, insert=False)
    assert not dao.exists(model)


def test_get__single_column(daos: TestDAOFactory):
    dao = daos[Student]
    dao.create_with(id=100, name='Bob')
    with daos.session_factory() as fresh_session:
        model = DAO(Student, fresh_session).get(100)
        assert model.id == 100
        assert model.name == 'Bob'


def test_get__multiple_column(daos: TestDAOFactory):
    dao = daos[Person]
    dao.create_with(name='John', age=23)
    with daos.session_factory() as fresh_session:
        model = DAO(Person, fresh_session).get('John', 23)
        assert model.name == 'John'
        assert model.age == 23


def test_get__missing_column(daos: TestDAOFactory):
    with pytest.raises(InvalidArgumentCount) as error:
        daos[Person].get('John')
    assert 'Expected 2 values, got 1' in str(error.value.detail)


def test_get__too_many_args(daos: TestDAOFactory):
    with pytest.raises(InvalidArgumentCount) as error:
        daos[Student].get(100, 'extra', 'args')
    assert 'Expected 1 values, got 3' in str(error.value.detail)


def test_get__not_found(daos: TestDAOFactory):
    with pytest.raises(NotFound):
        daos[Student].get(100)


def test_get_with(daos: TestDAOFactory):
    dao = daos[Student]
    dao.create_with(id=100, name='Bob')
    with daos.session_factory() as fresh_session:
        model = DAO(Student, fresh_session).get_with(id=100, name='Mike')
        assert model.id == 100
        assert model.name == 'Mike'


def test_get_with__no_additional_data(daos: TestDAOFactory):
    dao = daos[Student]
    dao.create_with(id=100, name='Bob')
    with daos.session_factory() as fresh_session:
        model = DAO(Student, fresh_session).get_with(id=100)
        assert model.id == 100
        assert model.name == 'Bob'


def test_get_with__blank_data(daos: TestDAOFactory):
    dao = daos[Student]
    dao.create_with(id=100, name='Bob')
    with daos.session_factory() as fresh_session:
        model = DAO(Student, fresh_session).get_with(id=100, name=None)
        assert model.id == 100
        assert model.name is None


def test_get__missing_data(daos: TestDAOFactory):
    with pytest.raises(MissingInput):
        daos[Person].get_with(name='John')


def test_get_with__not_found(daos: TestDAOFactory):
    with pytest.raises(NotFound):
        daos[Student].get_with(id=100, name='Mike')


def test_remove(daos: TestDAOFactory):
    dao = daos[Student]
    model = dao.create(100)
    daos.assert_in_db(Student, 100)
    dao.remove(model)
    daos.assert_not_in_db(Student, 100)


def test_remove__not_found(daos: TestDAOFactory):
    with pytest.raises(NotFound):
        daos[Student].remove(Student(id=100))


def test_commit(daos: TestDAOFactory):
    dao = daos[Student]
    model = dao.create(100)
    model.active = False
    daos.assert_in_db(Student, 100, active=True)
    dao.commit()
    daos.assert_in_db(Student, 100, active=False)


def test_commit__refresh(daos: TestDAOFactory):
    dao = daos[Student]
    dao.create(100)
    dao.start_transaction()
    model = dao.create_with(id=next_id())
    dao.commit(model)
    assert model.id is 101


def test_commit__missing(daos: TestDAOFactory):
    dao = daos[Student]
    model = dao.create(100)
    dao.start_transaction()
    dao.remove(model)
    with pytest.raises(InvalidRequestError):
        dao.commit(model)


def test_transaction(daos: TestDAOFactory):
    dao = daos[Student]
    dao.start_transaction()
    dao.create_with(id=100, name='Billy')
    dao.create_with(id=101, name='Bob')

    dao.commit()
    daos.assert_in_db(Student, 100, name='Billy')
    daos.assert_in_db(Student, 101, name='Bob')


def test_transaction__rollback(daos: TestDAOFactory):
    dao = daos[Student]
    dao.start_transaction()
    dao.create_with(id=100, name='Billy')
    dao.rollback()
    daos.assert_not_in_db(Student, 100)

    dao.create_with(id=101, name='Bob')
    dao.commit()
    daos.assert_in_db(Student, 101, name='Bob')
    daos.assert_not_in_db(Student, 100)


def transaction_setup(daos: TestDAOFactory) -> tuple[DAO, DAO]:
    student_dao = daos[Student]
    book_dao = daos[Book]

    student_dao.start_transaction()

    student_dao.create_with(id=103, name='Charlene')
    student_dao.db.flush()
    book_dao.create_with(name='Physics', subject='Science', owner=103)

    return student_dao, book_dao


def test_transaction__multiple_daos(daos: TestDAOFactory):
    student_dao, book_dao = transaction_setup(daos)

    book_dao.commit()

    daos.assert_in_db(Student, 103, name='Charlene')
    daos.assert_in_db(Book, 'Physics', subject='Science', owner=103)


def test_transaction__multiple_daos__rollback(daos: TestDAOFactory):
    student_dao, book_dao = transaction_setup(daos)

    student_dao.rollback()
    book_dao.commit()  # There is no longer anything to commit

    daos.assert_not_in_db(Student, 103)
    daos.assert_not_in_db(Book, 'Physics')


def test_transaction__multiple_daos__error(daos: TestDAOFactory):
    student_dao, book_dao = transaction_setup(daos)

    try:
        student_dao.create_with(id=103, name='Duplicate')
        student_dao.commit()
    except Conflict:
        student_dao.rollback()

    book_dao.commit()  # There is no longer anything to commit

    daos.assert_not_in_db(Student, 103)
    daos.assert_not_in_db(Book, 'Physics')


def test_transaction__multiple_sessions(daos: TestDAOFactory):
    pytest.skip("This test does not work with in-memory database")


def test_check_pk_arguments__single_column(daos: TestDAOFactory):
    dao = daos[Student]
    result = dao._check_pk_arguments((100,))
    assert result == {'id': 100}


def test_check_pk_arguments__multi_column(daos: TestDAOFactory):
    dao = daos[Person]
    result = dao._check_pk_arguments(('John', 23))
    assert result == {'name': 'John', 'age': 23}


def test_check_pk_arguments__too_few_args(daos: TestDAOFactory):
    dao = daos[Student]
    with pytest.raises(InvalidArgumentCount) as error:
        dao._check_pk_arguments(())
    assert 'Expected 1 values, got 0' in str(error.value.detail)
    assert 'Student primary key' in str(error.value.detail)


def test_check_pk_arguments__too_many_args(daos: TestDAOFactory):
    dao = daos[Student]
    with pytest.raises(InvalidArgumentCount) as error:
        dao._check_pk_arguments((100, 'extra', 'args'))
    assert 'Expected 1 values, got 3' in str(error.value.detail)
    assert 'Student primary key' in str(error.value.detail)



from datetime import date

import pytest

from daomodel.base_service import SingleModelService, SOURCE_VALUE, DESTINATION_VALUE, BaseService
from tests.conftest import TestDAOFactory
from tests.school_models import Staff, Student


def longest(values: list[str]) -> str:
    return max(values, key=len)


def setup_staff(daos: TestDAOFactory) -> tuple[Staff, Staff]:
    dao = daos[Staff]
    ed = dao.create_with(id=1, name='Ed', hire_date=date(2023, 6, 15))
    edward = dao.create_with(id=2, name='Edward', hire_date=date(2024, 8, 20))
    return ed, edward


def test_merge(daos: TestDAOFactory):
    ed, edward = setup_staff(daos)
    SingleModelService(daos, Staff).merge(ed, 2, name=longest, hire_date=min)
    daos.assert_in_db(Staff, 2, name='Edward', hire_date=date(2023, 6, 15))
    daos.assert_not_in_db(Staff, 1)


def test_merge__source_destination_values(daos: TestDAOFactory):
    ed, edward = setup_staff(daos)
    service = SingleModelService(daos, Staff)
    service.merge(edward, 1, name=DESTINATION_VALUE, hire_date=SOURCE_VALUE)
    daos.assert_in_db(Staff, 1, name='Ed', hire_date=date(2024, 8, 20))
    daos.assert_not_in_db(Staff, 2)


def test_merge__mismatched_model_type(daos: TestDAOFactory):
    setup_staff(daos)
    service = SingleModelService(daos, Staff)
    student = daos[Student].create_with(id=100, name='Student', gender='m')

    with pytest.raises(TypeError):
        service.merge(student, 1)


def test_dao(daos: TestDAOFactory):
    service = SingleModelService(daos, Staff)
    staff = service.dao.create_with(id=3, name='Alice', hire_date=date(2023, 1, 15))

    daos.assert_in_db(Staff, 3, name='Alice', hire_date=date(2023, 1, 15))

    staff.name = 'Alicia'
    service.dao.commit(staff)

    daos.assert_in_db(Staff, 3, name='Alicia', hire_date=date(2023, 1, 15))


def test_bulk_update(daos: TestDAOFactory):
    # Create multiple staff members
    dao = daos[Staff]
    staff1 = dao.create_with(id=10, name='Staff1', hire_date=date(2023, 1, 15))
    staff2 = dao.create_with(id=11, name='Staff2', hire_date=date(2023, 2, 20))
    staff3 = dao.create_with(id=12, name='Staff3', hire_date=date(2023, 3, 25))

    # Update all staff members with the same values
    service = SingleModelService(daos, Staff)
    service.bulk_update([staff1, staff2, staff3], hire_date=date(2024, 1, 1))

    # Verify all staff members were updated
    daos.assert_in_db(Staff, 10, name='Staff1', hire_date=date(2024, 1, 1))
    daos.assert_in_db(Staff, 11, name='Staff2', hire_date=date(2024, 1, 1))
    daos.assert_in_db(Staff, 12, name='Staff3', hire_date=date(2024, 1, 1))


def test_bulk_update__no_model_class(daos: TestDAOFactory):
    # Create multiple staff members
    dao = daos[Staff]
    staff1 = dao.create_with(id=20, name='Staff1', hire_date=date(2023, 1, 15))
    staff2 = dao.create_with(id=21, name='Staff2', hire_date=date(2023, 2, 20))

    # Update staff members without specifying model_class
    service = BaseService(daos)
    service.bulk_update([staff1, staff2], name='Updated Staff')

    # Verify all staff members were updated
    daos.assert_in_db(Staff, 20, name='Updated Staff', hire_date=date(2023, 1, 15))
    daos.assert_in_db(Staff, 21, name='Updated Staff', hire_date=date(2023, 2, 20))


def test_bulk_update__empty_list(daos: TestDAOFactory):
    # Should not raise an error when the list is empty
    service = SingleModelService(daos, Staff)
    service.bulk_update([], name='Should not be applied')


def test_bulk_update__mixed_model_types(daos: TestDAOFactory):
    # Create staff and student
    staff = daos[Staff].create_with(id=30, name='Staff', hire_date=date(2023, 1, 15))
    student = daos[Student].create_with(id=200, name='Student', gender='m')

    # Update models of different types
    service = BaseService(daos)
    service.bulk_update([staff, student], name='Updated Name')

    # Verify both models were updated
    daos.assert_in_db(Staff, 30, name='Updated Name', hire_date=date(2023, 1, 15))
    daos.assert_in_db(Student, 200, name='Updated Name', gender='m')


def test_bulk_update__multiple_mixed_models(daos: TestDAOFactory):
    # Create multiple staff and student models
    staff_dao = daos[Staff]
    student_dao = daos[Student]

    staff1 = staff_dao.create_with(id=40, name='Staff1', hire_date=date(2023, 1, 15))
    staff2 = staff_dao.create_with(id=41, name='Staff2', hire_date=date(2023, 2, 20))
    student1 = student_dao.create_with(id=300, name='Student1', gender='m')
    student2 = student_dao.create_with(id=301, name='Student2', gender='f')

    # Mix the models in a single list
    mixed_models = [staff1, student1, staff2, student2]

    # Update all models with a common field they inherit from BasePerson
    service = BaseService(daos)
    service.bulk_update(mixed_models, name='Common Name')

    # Verify all models were updated correctly
    daos.assert_in_db(Staff, 40, name='Common Name', hire_date=date(2023, 1, 15))
    daos.assert_in_db(Staff, 41, name='Common Name', hire_date=date(2023, 2, 20))
    daos.assert_in_db(Student, 300, name='Common Name', gender='m')
    daos.assert_in_db(Student, 301, name='Common Name', gender='f')



from typing import Any

from daomodel.util import names_of


class PropertyFilter:
    """Base class for property filters with operator overloading support."""
    
    def evaluate(self, model: Any) -> set[str]:
        """Evaluates this filter against the model and returns matching property names.
        
        :param model: The model to evaluate against
        :return: A set of property names that match this filter
        """
        raise NotImplementedError('Must implement `evaluate` in subclass')
    
    def __and__(self, other):
        """Overload the & operator to create an AND filter."""
        return AndFilter(self, other)
    
    def __or__(self, other):
        """Overload the | operator to create an OR filter."""
        return OrFilter(self, other)
    
    def __invert__(self):
        """Overload the ~ operator to create a NOT filter."""
        return NotFilter(self)


class AndFilter(PropertyFilter):
    """Combines multiple filters with AND logic."""
    
    def __init__(self, left, right):
        self.left = left
        self.right = right
    
    def evaluate(self, model) -> set[str]:
        """Returns properties that match ALL of the combined filters."""
        return self.left.evaluate(model).intersection(self.right.evaluate(model))
    
    def __repr__(self):
        return f'({self.left} & {self.right})'


class OrFilter(PropertyFilter):
    """Combines multiple filters with OR logic."""
    
    def __init__(self, left, right):
        self.left = left
        self.right = right
    
    def evaluate(self, model) -> set[str]:
        """Returns properties that match ANY of the combined filters."""
        return self.left.evaluate(model).union(self.right.evaluate(model))
    
    def __repr__(self):
        return f'({self.left} | {self.right})'


class NotFilter(PropertyFilter):
    """Negates a filter."""
    
    def __init__(self, operand):
        self.operand = operand
    
    def evaluate(self, model) -> set[str]:
        """Returns properties that do NOT match the given filter."""
        all_props = set(names_of(model.get_properties()))
        return all_props.difference(self.operand.evaluate(model))
    
    def __repr__(self):
        return f'~{self.operand}'


class BasicPropertyFilter(PropertyFilter):
    """Filter for specific property categories."""
    
    def __init__(self, name):
        self.name = name
    
    def evaluate(self, model) -> set[str]:
        """Returns properties that match this category."""
        if self.name == 'ALL':
            return set(names_of(model.get_properties()))
        elif self.name == 'PK':
            return set(model.get_pk_names())
        elif self.name == 'FK':
            return set(names_of(model.get_fk_properties()))
        elif self.name == 'DEFAULT':
            return ALL.evaluate(model).difference(model.model_dump(exclude_defaults=True))
        elif self.name == 'NONE':
            return ALL.evaluate(model).difference(model.model_dump(exclude_none=True))
        return set()
    
    def __repr__(self):
        return f'{self.name}'


ALL = BasicPropertyFilter('ALL')
PK = BasicPropertyFilter('PK')
FK = BasicPropertyFilter('FK')
DEFAULT = BasicPropertyFilter('DEFAULT')
NONE = BasicPropertyFilter('NONE')



from typing import Any, Callable, List

from daomodel import DAOModel, all_models
from daomodel.db import DAOFactory
from daomodel.model_diff import ChangeSet, Preference


SOURCE_VALUE = Preference.RIGHT
DESTINATION_VALUE = Preference.LEFT


class BaseService:
    """An extendable base for creating a service layer for your DAOModel project.

    This service provides access to DAOs for model classes through the `daos` property.
    It also provides methods for operations involving your models.
    """
    def __init__(self, daos: DAOFactory):
        self.daos = daos

    def bulk_update(self, models: List[DAOModel], **common_values: Any) -> None:
        """Assigns values to multiple models in a single transaction.

        This is particularly useful for applying batch changes to a filtered set of records.
        ```python
        service.bulk_update(overdue_accounts, status='suspended')
        ```

        The method supports updating models of different types in a single operation,
        as long as the fields being updated are present in all model types.
        ```python
        no_longer_at_school = (school_service.daos[Student].find(graduated=True) +
                               school_service.daos[Staff].find(retired=True))
        service.bulk_update(no_longer_at_school, status='inactive', door_code=None)
        ```

        :param models: List of models to update (can be of different types)
        :param common_values: Values to set on all models
        """
        if not models:
            return

        # Group models by type
        models_by_type = {}
        for model in models:
            model_type = type(model)
            if model_type not in models_by_type:
                models_by_type[model_type] = []
            models_by_type[model_type].append(model)

        # Start a transaction on the first model type's DAO
        # This will be used for all model types
        if not models_by_type:
            return

        first_model_type = next(iter(models_by_type.keys()))
        transaction_dao = self.daos[first_model_type]
        transaction_dao.start_transaction()

        try:
            # Process each model type
            for model_type, model_list in models_by_type.items():
                model_dao = self.daos[model_type]
                for model in model_list:
                    model.set_values(**common_values)
                    model_dao.db.add(model)

            # Commit the transaction
            transaction_dao.commit()
        except Exception as e:
            transaction_dao.rollback()
            raise e


class SingleModelService(BaseService):
    """A service layer specifically designed around a single DAOModel.

    This service extends BaseService and provides access to the DAO for the primary model 
    through the `dao` property, allowing direct access to all DAO methods without creating 
    pass-through methods. It also provides additional methods that add value beyond simple 
    DAO operations.
    """
    def __init__(self, daos: DAOFactory, model_class: type[DAOModel]):
        super().__init__(daos)
        self.dao = daos[model_class]

    def merge(self, source: DAOModel, *destination_pk_values, **conflict_resolution: Preference|Callable|Any) -> None:
        """Merges the given source model into the specified destination.

        In some cases, specify conflict_resolution to successfully merge values.
        See the `ChangeSet` documentation for more details on conflict resolution.

        :param source: The source DAOModel to be merged into the destination
        :param destination_pk_values: The primary key values indicating where to merge the model
        :raises NotFound: if the destination model does not exist in the database
        :raises Conflict: if the source model fails to merge into the destination
        :raises TypeError: if source is not of the primary model type
        """
        model_dao = self.dao
        model_dao.start_transaction()
        destination = model_dao.get(*destination_pk_values)
        if type(source) is not self.dao.model_class:
            raise TypeError(f'{source} is not of type {self.dao.model_class}')

        ChangeSet(destination, source, **conflict_resolution).resolve_preferences().apply()
        self._redirect_fks(source, destination)
        model_dao.remove(source)
        model_dao.commit(destination)

    def _redirect_fks(self, source: DAOModel, destination: DAOModel) -> None:
        for model in all_models(self.daos.db.get_bind()):
            model_dao = self.daos[model]
            for column in model.get_references_of(source):
                old_value = source.get_value_of(column)
                new_value = destination.get_value_of(column)
                model_dao.query.where(column == old_value).update({column: new_value})



import inspect
from dataclasses import dataclass
from typing import Callable, TypeVar, Generic

import pytest


T = TypeVar('T')

@dataclass
class Expected(Generic[T]):
    """Wrapper for expected test results to be shared across multiple test cases"""
    value: T

TestCase = tuple
TestGroup = list[Expected | TestCase]
TestCases = TestGroup | TestCase


def _validate_parameters(param_count: int, test_case: TestCase) -> None:
    if param_count > 1:
        if not isinstance(test_case, tuple):
            raise ValueError(f'Expected tuple of {param_count} parameters but got {test_case}')
        elif len(test_case) != param_count:
            raise ValueError(f'Expected {param_count} parameters but got {len(test_case)}: ({test_case})')


def labeled_tests(tests: dict[str, TestCases]):
    def decorator(test_func: Callable):
        parameter_names = inspect.signature(test_func).parameters.keys()
        param_count = len(parameter_names)
        labels = []
        test_data = []

        for group_label, test_cases in tests.items():
            if not isinstance(test_cases, list):
                test_cases = [test_cases]

            expected = None
            try:
                expected_case = next(case for case in test_cases if isinstance(case, Expected))
                expected = expected_case.value
                test_cases.remove(expected_case)
                found_expected = True
            except StopIteration:
                found_expected = False

            for test_case in test_cases if isinstance(test_cases, list) else [test_cases]:
                if found_expected:
                    test_case = (test_case, expected)
                _validate_parameters(param_count, test_case)
                labels.append(group_label)
                test_data.append(test_case)
        return pytest.mark.parametrize(', '.join(parameter_names), test_data, ids=labels)(test_func)

    return decorator



from typing import Any, Optional
from uuid import UUID

import pytest
from sqlalchemy.exc import IntegrityError

from daomodel.fields import Protected, Identifier, Unsearchable
from daomodel.util import ensure_iter, names_of
from tests.conftest import TestDAOFactory
from tests.labeled_tests import labeled_tests
from tests.model_factory import create_test_model
from tests.test_fields import BasicModel


all_test_cases = {
    'standard field': str,
    'uuid': UUID,
    'dict': dict,
    'reference': BasicModel,
    'protected reference': Protected[BasicModel],
    'identifier field': Identifier[str],
    'identifier uuid': Identifier[UUID],
    # Note: dict cannot be an Identifier as it is not hashable
    'identifier reference': Identifier[BasicModel],
    'identifier protected reference': Identifier[Protected[BasicModel]],
    'optional field': Optional[str],
    'optional uuid': Optional[UUID],
    'optional dict': Optional[dict],
    'optional reference': Optional[BasicModel],
    'optional protected reference': Protected[Optional[BasicModel]],
    'unsearchable field': Unsearchable[str],
    'unsearchable uuid': Unsearchable[UUID],
    'unsearchable dict': Unsearchable[dict],
    'unsearchable reference': Unsearchable[BasicModel],
    'unsearchable protected reference': Unsearchable[Protected[BasicModel]],
    'identifier + optional field': Identifier[Optional[str]],
    'identifier + optional uuid': Identifier[Optional[UUID]],
    # Note: dict cannot be an Identifier as it is not hashable
    'identifier + optional reference': Identifier[Optional[BasicModel]],
    'identifier + optional protected reference': Identifier[Protected[Optional[BasicModel]]],
    'unsearchable + identifier field': Unsearchable[Identifier[str]],
    'unsearchable + identifier uuid': Unsearchable[Identifier[UUID]],
    # Note: dict cannot be an Identifier as it is not hashable
    'unsearchable + identifier reference': Unsearchable[Identifier[BasicModel]],
    'unsearchable + identifier protected reference': Unsearchable[Identifier[Protected[BasicModel]]],
    'unsearchable + optional field': Unsearchable[Optional[str]],
    'unsearchable + optional uuid': Unsearchable[Optional[UUID]],
    'unsearchable + optional dict': Unsearchable[Optional[dict]],
    'unsearchable + optional reference': Unsearchable[Optional[BasicModel]],
    'unsearchable + optional protected reference': Unsearchable[Protected[Optional[BasicModel]]],
    'unsearchable + identifier + optional field': Unsearchable[Identifier[Optional[str]]],
    'unsearchable + identifier + optional uuid': Unsearchable[Identifier[Optional[UUID]]],
    # Note: dict cannot be an Identifier as it is not hashable
    'unsearchable + identifier + optional reference': Unsearchable[Identifier[Optional[BasicModel]]],
    'unsearchable + identifier + optional protected reference': Unsearchable[Identifier[Protected[Optional[BasicModel]]]]
}


def get_test_cases(*labels: str, exclude: str|tuple[str, ...] = ()) -> dict:
    return {k: v for k, v in all_test_cases.items()
            if all(label in k for label in labels) and not any(ex in k for ex in ensure_iter(exclude))}


@labeled_tests(get_test_cases('field'))
def test_inherited_field(annotation: Any):
    model_type = create_test_model(annotation, inherited=True)
    with TestDAOFactory() as daos:
        dao = daos[model_type]
        dao.create_with(id=1, value='test', child_field='extended')
        entry = dao.find(id=1)
        assert entry.only().value == 'test'
        assert entry.only().child_field == 'extended'


@labeled_tests(get_test_cases('uuid'))
def test_inherited_uuid(annotation: Any):
    model_type = create_test_model(annotation, inherited=True)
    with TestDAOFactory() as daos:
        dao = daos[model_type]
        entry = dao.create_with(id=1, child_field='extended')

        assert getattr(entry, 'value') is not None
        assert isinstance(getattr(entry, 'value'), UUID)

        entry2 = dao.create_with(id=2, child_field='extended')
        assert getattr(entry2, 'value') is not None
        assert isinstance(getattr(entry2, 'value'), UUID)
        assert getattr(entry, 'value') != getattr(entry2, 'value')


@labeled_tests(get_test_cases('dict'))
def test_inherited_dict(annotation: Any):
    model_type = create_test_model(annotation, inherited=True)
    sample_dict = {'Hello': 'World!', 'Lorem': ['Ipsum', 'Dolor', 'Sit', 'Amet']}
    with TestDAOFactory() as daos:
        daos[model_type].create_with(id=1, value=sample_dict, child_field='extended')
        daos.assert_in_db(model_type, 1, value=sample_dict, child_field='extended')


@labeled_tests(get_test_cases('reference'))
def test_inherited_reference(annotation: Any):
    model_type = create_test_model(annotation, inherited=True)
    with TestDAOFactory() as daos:
        daos[BasicModel].create(100)
        dao = daos[model_type]
        dao.create_with(id=1, value=100, child_field='extended')
        entry = dao.find(id=1)
        assert entry.only().value == 100
        assert entry.only().child_field == 'extended'


@labeled_tests(get_test_cases(exclude=('optional', 'uuid')))
def test_inherited_required(annotation: Any):
    model_type = create_test_model(annotation, inherited=True)
    with TestDAOFactory() as daos:
        dao = daos[model_type]
        entry = dao.create_with(id=1, child_field='extended', insert=False)
        with pytest.raises(IntegrityError):
            dao.insert(entry)


@labeled_tests(get_test_cases('uuid', exclude='optional'))
def test_inherited_required__uuid(annotation: Any):
    model_type = create_test_model(annotation, inherited=True)
    with TestDAOFactory() as daos:
        dao = daos[model_type]
        dao.create_with(id=1, child_field='extended')
        entry = dao.find(id=1).only()
        # Since UUID has a default value, we test that it is Required by attempting to set it to None
        entry.value = None
        with pytest.raises(IntegrityError):
            dao.upsert(entry)


@labeled_tests(get_test_cases('optional', exclude='uuid'))
def test_inherited_optional(annotation: Any):
    model_type = create_test_model(annotation, inherited=True)
    with TestDAOFactory() as daos:
        dao = daos[model_type]
        dao.create_with(id=1, child_field='extended')
        entry = dao.find(id=1).only()
        assert entry.value is None


@labeled_tests(get_test_cases('optional', 'uuid'))
def test_inherited_optional__uuid(annotation: Any):
    model_type = create_test_model(annotation, inherited=True)
    with TestDAOFactory() as daos:
        dao = daos[model_type]
        dao.create_with(id=1, child_field='extended')
        entry = dao.find(id=1).only()
        assert entry.value is not None
        # Since UUID has a default value, we test that it is Optional by setting it to None
        entry.value = None
        dao.upsert(entry)
        entry = dao.find(id=1).only()
        assert entry.value is None


@labeled_tests(get_test_cases('identifier'))
def test_inherited_identifier(annotation: Any):
    assert create_test_model(annotation, field_name='id2', inherited=True).get_pk_names() == ['id', 'id2']


@labeled_tests(get_test_cases(exclude='unsearchable'))
def test_inherited_searchable(annotation: Any):
    assert names_of(create_test_model(annotation, inherited=True).get_searchable_properties()) == ['id', 'value', 'child_field']


@labeled_tests(get_test_cases('unsearchable'))
def test_inherited_unsearchable(annotation: Any):
    assert names_of(create_test_model(annotation, inherited=True).get_searchable_properties()) == ['id', 'child_field']
